# Copyright Okera Inc.

from __future__ import absolute_import

import collections
import datetime
import functools
import hashlib
import heapq
import json
import time
import sys

from okera._thrift_api import TAuthorizeQueryClient
from okera._thrift_api import TAuthorizeQueryParams
from okera._thrift_api import TGetDatasetsParams
from okera._thrift_api import TErrorCode
from okera._thrift_api import TRecordServiceException

def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

# Value to use in masking policy when the user does not have access to the column
NO_ACCESS_VALUE = 'NULL'

SNOWFLAKE_ROW_POLICY_PREAMBLE = """
CREATE OR REPLACE ROW ACCESS POLICY %s AS (%s) RETURNS boolean ->
"""

SNOWFLAKE_MASKING_POLICY_PREAMBLE = """
CREATE OR REPLACE MASKING POLICY %s AS (%s %s) RETURNS %s ->
"""

# These are (partial) statements that are okay to fail. Snowflake doesn't accept
# 'IF NOT EXISTS' in these statements so we have to try ignore errors.
ACCEPTABLE_ERROR_QUERIES = [
  'DROP ROW ACCESS POLICY',
  'DROP MASKING POLICY',
]

# And expected errors for those policies
APPLY_POLICY_NOT_EXISTS_SENTINEL1 = 'does not exist or not authorized.'
APPLY_POLICY_NOT_EXISTS_SENTINEL2 = 'is not attached to'

# Table property to control if syncing is enabled
TBL_PROP_DRIVER = 'driver'
TBL_PROP_SYNC_ENABLED = 'okera.policy-sync.enabled'

# Table properties to control the users to sync
TBL_PROP_USERS_ADDITIONAL = 'okera.policy-sync.users.additional'
TBL_PROP_USERS_BLACKLIST = 'okera.policy-sync.users.blacklist'
TBL_PROP_USERS_LIST = 'okera.policy-sync.users.list'
TBL_PROP_USERS_WHITELIST = 'okera.policy-sync.users.whitelist'

# Table properties for policy sync status
# Keep in sync with PolicySync.java
TBL_PROP_POLICY_LAST_ERROR = "policy-sync.last-syncing-utc"
TBL_PROP_POLICY_LAST_QUEUED = "policy-sync.last-queued-utc"
TBL_PROP_POLICY_LAST_SYNC = "policy-sync.last-sync-utc"
TBL_PROP_POLICY_LAST_SYNCING = "policy-sync.last-syncing-utc"
TBL_PROP_POLICY_SYNC_MD5 = "policy-sync.md5"
TBL_PROP_POLICY_SYNC_ERROR_MSG = "policy-sync.error-message"
TBL_PROP_POLICY_SYNC_STATUS_MSG = "policy-sync.status-message"

def _masking_policy_name(tbl, col):
    return "%s_%s_masking_policy" % (tbl, col)

def _row_policy_name(tbl):
    return tbl + '_row_policy'

class CoalescingEventQueue():
    """ A coalescing producer/consumer event queue. The only difference is it adds
        coalescing. This means that if the same key is enqueue in a short window, only
        one (the last) event will be dequeued. This is used typically to model a
        triggering pattern and only the last event in a burst should trigger.

        This is done by delaying when a enqueued event can be dequeued and then
        scanning the queue for other events with the same event at dequeue time. If
        another of the same event is found, the earlier event is ignored (removed from
        queue and not returned.
    """
    def __init__(self, delay_ms, max_ms):
        """ delay_ms is the time to wait before an enqueue is visible for
            dequeue. It is the time to wait for coalescing.

            max_ms is the maximum time to wait before an event is triggered
            and will not longer be eligible for coalescing.
        """
        self.delay_ms = delay_ms
        self.max_ms = max_ms
        self.heap = []

    def enqueue(self, e):
        _enqueue(e, __now())

    def dequeue(self):
        return self._dequeue(__now())

    # Test helpers to control the clock.
    def _enqueue(self, e, ts):
        heapq.heappush(self.heap, (ts + self.delay_ms, e, ts))

    def _dequeue(self, now):
        print(self.heap)
        while self.heap:
            v = self.heap[0]
            if self.max_ms and v[2] + self.max_ms < now:
                # Handle max_ms
                heappq.heappop(self.heap)
                return v[1]

            if v[0] > now:
                return None

            heapq.heappop(self.heap)
            for e in self.heap:
                if e[1] == v[1]:
                    # Found another entry, ignore this one
                    v = None
                    break
            if not v:
                continue
            return v[1]

        return None

    def __now(self):
        return round(time.timme() * 1000)

class Policy():
    def __init__(self, user):
        self.user = user
        self.filters = []

        # Empty indicates access to all columns
        self.column_restrictions = {}

    def __repr__(self):
        return 'User: %s. Filters: %s. Columns: %s' % \
            (self.user, self.filters, self.column_restrictions)

def _compute_hash(*args):
    s = ''
    for a in args:
        s += str(a)
    return hashlib.md5(s.encode('utf-8')).hexdigest()

def _generate_grants(
        users, policies, single_grant_role, per_user_role_grants, tbl, role_pattern):
    result = []
    roles = []
    if single_grant_role:
        roles.append(single_grant_role)
        for u, p in policies.items():
            if p:
                result.append('GRANT SELECT ON %s TO ROLE %s' % \
                    (tbl, single_grant_role))
            else:
                result.append('REVOKE SELECT ON %s FROM ROLE %s' % \
                    (tbl, single_grant_role))
    elif per_user_role_grants:
        for u, p in policies.items():
            role = role_pattern % u
            roles.insert(0, role)
            if p:
                result.append('GRANT SELECT ON %s TO ROLE %s' % (tbl, role))
                namespaces = tbl.split('.')
                # TODO relax this
                assert len(namespaces) == 3
                result.append(
                    'GRANT USAGE ON DATABASE %s TO ROLE %s' % (namespaces[0], role))
                result.append(
                    'GRANT USAGE ON SCHEMA %s.%s TO ROLE %s' % (
                        namespaces[0], namespaces[1], role))
            else:
                result.append('REVOKE SELECT ON %s FROM ROLE %s' % (tbl, role))
            if u in users:
                result.append('GRANT ROLE %s TO USER "%s"' % (role, u))
    for role in roles:
        result.insert(0, 'CREATE ROLE IF NOT EXISTS %s' % role)
    return result

def _none_safe_cmp(x, y):
    if isinstance(x, tuple):
        for i in range(0, len(x)):
            v = _none_safe_cmp(x[i], y[i])
            if v:
                return v
        return 0
    if x is None or y is None:
        if x is None and y is None:
            return 0
        elif x is None:
            return -1
        else:
            return 1
    else:
        if x == y:
            return 0
        elif x < y:
            return -1
        else:
            return 1

def _generate_row_filters(policies, tbl, column_signatures):
    need_filter_policy = False
    # Map of filter to users with that filter. Empty string indicates no filter
    filter_policies = {}
    for u, p in policies.items():
        if not p:
            need_filter_policy = True
            continue
        k = None
        if not p or not p.filters:
            k = ''
        else:
            need_filter_policy = True
            k = ' AND '.join(p.filters)

        if k not in filter_policies:
            filter_policies[k] = []
        filter_policies[k].append(u)

    if not need_filter_policy or not column_signatures:
        return None, None

    policy_name = _row_policy_name(tbl)

    # For each filter, generate the case statement
    policy = SNOWFLAKE_ROW_POLICY_PREAMBLE % (policy_name, ', '.join(column_signatures))
    policy += 'CASE\n'
    for p, u in filter_policies.items():
        users = []
        for user in u:
            users.append("'%s'" % user)
        if not p:
            p = 'TRUE'
        case = '    WHEN current_user() in (%s) THEN %s\n' % (', '.join(users), p)
        policy += case
    policy += '    ELSE FALSE\n'
    policy += 'END'

    return policy, policy_name

def _generate_column_policies(policies, tbl):
    # Map of (col -> map of(transform -> list of users)
    partial_access_users = {}
    no_access_users = []
    all_access_users = []

    referenced_columns = {}

    for u, p in policies.items():
        if not p:
            no_access_users.append("'%s'" % u)
            continue
        if not p.column_restrictions:
            all_access_users.append("'%s'" % u)
            continue

        for col, access in p.column_restrictions.items():
            if col not in partial_access_users:
                partial_access_users[col] = {}

            if access.transform:
                v = access.transform
            elif access.accessible:
                v = col
            else:
                v = None

            if v not in partial_access_users[col]:
                partial_access_users[col][v] = []
            partial_access_users[col][v].append(u)
            referenced_columns[col] = access.column

    if not partial_access_users:
        # No partial access, no need to produce column policy
        return None

    partial_access_users = collections.OrderedDict(
        sorted(partial_access_users.items(), key=functools.cmp_to_key(_none_safe_cmp)))

    # Collection of masking policies. This will be populated for each column
    # that needs a masking policy.
    masking_policies = []
    for col, p in partial_access_users.items():
        col_desc = referenced_columns[col]
        policy_name = _masking_policy_name(tbl, col)
        policy = SNOWFLAKE_MASKING_POLICY_PREAMBLE % \
            (policy_name, col_desc.name, col_desc.type, col_desc.type)
        policy += 'CASE\n'

        # Add the cases for each type of transform
        p = collections.OrderedDict(
            sorted(p.items(), key=functools.cmp_to_key(_none_safe_cmp)))

        # Handle users with full access
        full_access_users = all_access_users.copy()
        for t, u in p.items():
            if t == col:
                for user in u:
                    full_access_users.append("'%s'" % user)
        case = '    WHEN current_user() in (%s) THEN %s\n' % \
            (', '.join(full_access_users), col)
        policy += case

        # In the case where the only policies with all users having full access,
        # no need to create a column policy.
        need_to_add = False

        # 't' is the target transformed value and u is the list of users needing
        # this transform.
        for t, u in p.items():
            # v is the expression to transform to
            if t == col:
                # Handled above with all_access_users
                continue

            need_to_add = True
            v = t
            if t is None:
                v = NO_ACCESS_VALUE

            # Collect all the users that need this transform on this column
            users = []
            for user in u:
                users.append("'%s'" % user)
            case = '    WHEN current_user() in (%s) THEN %s\n' % \
                (', '.join(users), v)
            policy += case

        if need_to_add:
            policy += '    ELSE %s\n' % NO_ACCESS_VALUE
            policy += 'END'
            masking_policies.append((col, policy_name, policy))
    return masking_policies

def debug_print(grants, pre_alters, row_filters, column_policies, post_alters):
    print("GRANTS:")
    print("--------------------------------------------------------------")
    for grant in grants:
        print(grant + ';')
    print("--------------------------------------------------------------")

    print("ROW FILTER:")
    print("--------------------------------------------------------------")
    if row_filters:
        print(row_filters + ';')
    print("--------------------------------------------------------------")

    print("COLUMN POLICIES:")
    print("--------------------------------------------------------------")
    if column_policies:
        for p in column_policies:
            print(p[2] + ';')
    print("--------------------------------------------------------------")

    print("PRE-ALTERS:")
    print("--------------------------------------------------------------")
    for alter in pre_alters:
        print(alter + ';')
    print("--------------------------------------------------------------")

    print("POST-ALTERS:")
    print("--------------------------------------------------------------")
    for alter in post_alters:
        print(alter + ';')
    print("--------------------------------------------------------------")

def generate_policies(
        conn, users, db_name, tbl_name, okera_tbl,
        single_grant_role=None,
        per_user_role_grants=True,
        role_pattern='"%s_ROLE"',
        client=TAuthorizeQueryClient.SNOWFLAKE):
    """ Generates the set of DDL needed to update this table for all users

        This does not mutate the underlying database at all. It only generates
        the required SQL and does not run them.

        Parameters
        ----------
        conn :
            Okera planner connection.

        users : list[str]
            Total list of users to generate policies for.

        db_name : str
            Database, in the okera catalog to generate policies for.

        okera_tbl : obj
            Okera table object

        tbl_name : str
            Table in database, in the okera catalog to generate policies for.

        Returns
        -------
        Tuple of different types of DDL to run.
    """

    if client != TAuthorizeQueryClient.SNOWFLAKE:
        raise RuntimeError("Only SNOWFLAKE is a supported client currently.")

    input_users = users.copy()
    # Adding system user to this list to ensure we can get metadata back
    if 'okera' not in users:
        users.append('okera')

    # For every user, look up the policies and populate 'policies'
    # Users that have no access will have None as the map value
    policies = {}

    # Columns and types used across all filters across all users
    row_filter_column_names = []
    row_filter_column_signatures = []

    # Table name in snowflake
    sf_table_name = None
    fq_tbl_name = db_name + '.' + tbl_name

    for user in users:
        request = TAuthorizeQueryParams()
        request.client = TAuthorizeQueryClient.SNOWFLAKE_POLICY_SYNC
        request.db = [db_name]
        request.dataset = tbl_name
        request.requesting_user = user

        try:
            result = conn.service.client.AuthorizeQuery(request)
            p = Policy(user)
            policies[user] = p

            if result.jdbc_referenced_tables and \
                    fq_tbl_name in result.jdbc_referenced_tables:
                sf_table_name = result.jdbc_referenced_tables[fq_tbl_name].jdbc_fq_tbl_name

            if result.filters:
                p.filters += result.filters.filters

                for col in result.filters.columns:
                    if col.name in row_filter_column_names:
                        # Already seen this column.
                        continue
                    row_filter_column_names.append(col.name)
                    row_filter_column_signatures.append('%s %s' % (col.name, col.type))

            if result.column_access:
                p.column_restrictions = result.column_access
        except TRecordServiceException as ex:
            if ex.code == TErrorCode.AUTHENTICATION_ERROR:
                policies[user] = None
            else:
                raise ex

    if not sf_table_name:
        raise RuntimeError("Could not determine Snowflake table name. Cannot sync.")

    sf_columns = []
    for col in okera_tbl.schema.cols:
        sf_columns.append(col.name.upper())

    # Sort the policies by user for stability
    policies = collections.OrderedDict(
        sorted(policies.items(), key=functools.cmp_to_key(_none_safe_cmp)))

    # Generate the grants
    grants = _generate_grants(
        input_users, policies, single_grant_role, per_user_role_grants, sf_table_name,
        role_pattern)

    # Generate the row filters
    row_filter_policy, row_filter_policy_name = _generate_row_filters(
        policies, sf_table_name, row_filter_column_signatures)

    # Generate the column masking policies
    column_policies = _generate_column_policies(
        policies, sf_table_name)

    # Generate the alters
    pre_alters = []
    post_alters = []

    # We need to reset the state of this table before we proceed, so
    # drop any previously set policies.
    # Note: some of these fail with a "not exists" error that should be ignored.
    pre_alters.append(
        'ALTER TABLE %s DROP ROW ACCESS POLICY %s' % \
        (sf_table_name, _row_policy_name(sf_table_name)))
    for sf_col in sf_columns:
      pre_alters.append(
          'ALTER TABLE %s MODIFY COLUMN %s UNSET MASKING POLICY' % \
          (sf_table_name, sf_col))
      pre_alters.append(
          'DROP MASKING POLICY %s' % (_masking_policy_name(sf_table_name, sf_col)))

    if row_filter_policy:
        post_alters.append(
            'ALTER TABLE %s ADD ROW ACCESS POLICY %s ON(%s)' % \
            (sf_table_name, row_filter_policy_name, ', '.join(row_filter_column_names)))
    if column_policies:
        for col, policy, _ in column_policies:
            post_alters.append(
                'ALTER TABLE %s MODIFY COLUMN %s SET MASKING POLICY %s' % \
                (sf_table_name, col, policy))

    md5 = _compute_hash(\
        grants, pre_alters, row_filter_policy, column_policies, post_alters);
    return grants, pre_alters, row_filter_policy, column_policies, post_alters, md5

def get_snowflake_users(conn, before_role=None, after_role=None):
    if before_role:
        conn.query_request("USE ROLE {}".format(before_role))
    result = conn.query_request("SHOW USERS")
    if after_role:
        conn.query_request("USE ROLE {}".format(after_role))
    users = []
    for u in result['rowset']:
        users.append(u[0])
    return users

def get_sync_users(users, tbl_params):
    """ Computes the users to sync combining the input users with the configs on
        the table. 'users' typically will come from the underlying db.
    """
    if not tbl_params:
        return users

    result = users.copy()

    def to_list(v):
        l = []
        for u in v.split(','):
            u = u.strip()
            if u:
                l.append(u)
        return l

    if TBL_PROP_USERS_LIST in tbl_params:
        # Set in table properties, use that
        v = tbl_params[TBL_PROP_USERS_LIST]
        result = []
        for u in to_list(v):
            result.append(u)
        return result

    if TBL_PROP_USERS_ADDITIONAL in tbl_params:
        # Add any additional users
        v = tbl_params[TBL_PROP_USERS_ADDITIONAL]
        for u in to_list(v):
            result.append(u)

    if TBL_PROP_USERS_BLACKLIST in tbl_params:
        # Remove any blacklist users
        v = tbl_params[TBL_PROP_USERS_BLACKLIST]
        for u in to_list(v):
            if u in result:
                result.remove(u)

    if TBL_PROP_USERS_WHITELIST in tbl_params:
        # Only return whitelist users
        v = tbl_params[TBL_PROP_USERS_WHITELIST]
        whitelist = []
        for u in to_list(v):
            if u in result:
                whitelist.append(u)
        return whitelist

    return result

def sync_policies(
        planner_conn, db_conn, db_name, tbl_name, role, users=None,
        role_pattern='"%s_ROLE"',
        per_user_role_grants=True,
        dry_run=False,
        client=TAuthorizeQueryClient.SNOWFLAKE):
    """ Synchronizes the policies for this table.

        Parameters
        ----------
        planner_conn :
            Okera planner connection.

        db_conn :
            Connection to the underlying DB.

        users : list[str], optional
            If set, the users to generate for, otherwise, determine them from the
            environment.

        db_name : str
            Database, in the okera catalog to generate policies for.

        tbl_name : str
            Table in database, in the okera catalog to generate policies for.
    """

    eprint('Syncing table %s.%s...' % (db_name, tbl_name))

    # Get the okera table object
    request = TGetDatasetsParams()
    request.dataset_names = ['{}.{}'.format(db_name, tbl_name)]
    request.with_schema = True
    okera_tbl = planner_conn.service.client.GetDatasets(request).datasets[0]

    # Update the last time a syncing operation begun
    now = datetime.datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
    properties = [
        "'%s'='%s'" % (TBL_PROP_POLICY_LAST_SYNCING, now),
    ]
    sql = 'ALTER TABLE `%s`.`%s` SET TBLPROPERTIES(%s)' % (db_name, tbl_name, ','.join(properties))
    if not dry_run:
        planner_conn.execute_ddl(sql)

    if not users:
        eprint('  Querying snowflake for users...')
        users = get_snowflake_users(db_conn, before_role=role)
    eprint('  Users: %s' % users)

    users = get_sync_users(users, okera_tbl.metadata)
    eprint('  Users after table configs: %s' % users)

    grants, pre_alters, row_filters, column_policies, post_alters, md5 = \
        generate_policies(planner_conn, users, db_name, tbl_name, okera_tbl,
                          per_user_role_grants=per_user_role_grants,
                          role_pattern=role_pattern)
    debug_print(grants, pre_alters, row_filters, column_policies, post_alters)

    def run(sql):
        eprint('...running: %s' % sql)
        if not dry_run:
            try:
                db_conn.query_request(sql)
            except Exception as e:
                if e.__str__().endswith(APPLY_POLICY_NOT_EXISTS_SENTINEL1) or \
                    APPLY_POLICY_NOT_EXISTS_SENTINEL2 in e.__str__():
                    for q in ACCEPTABLE_ERROR_QUERIES:
                        if q in sql:
                          eprint('......Skipping query with expected error.')
                          return
                eprint('Could not run DDL: %s' % sql)
                raise

    eprint('Running DDL against snowflake...')
    start = time.time()
    num_ddls = 0
    if grants:
        for g in grants:
            num_ddls += 1
            run(g)
    if pre_alters:
        for a in pre_alters:
            num_ddls += 1
            run(a)
    if row_filters:
        num_ddls += 1
        run(row_filters)
    if column_policies:
        for c in column_policies:
            num_ddls += 1
            run(c[2])
    if post_alters:
        for a in post_alters:
            num_ddls += 1
            run(a)
    eprint('Done updating snowflake(total ddl: %s) in %ss\n' %\
           (num_ddls, time.time() - start))

    eprint('Updating Okera catalog...')
    now = datetime.datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
    properties = [
        "'%s'='%s'" % (TBL_PROP_POLICY_SYNC_MD5, md5),
        "'%s'='%s'" % (TBL_PROP_POLICY_LAST_SYNC, now),
    ]
    sql = 'ALTER TABLE `%s`.`%s` SET TBLPROPERTIES(%s)' % (db_name, tbl_name, ','.join(properties))
    eprint('...Alter DDL: %s' % sql)
    if not dry_run:
        planner_conn.execute_ddl(sql)
    return grants, pre_alters, row_filters, column_policies, post_alters, md5

def sync_catalog(planner_conn, db_conn, db_role, stats, whitelist_dbs=None,
                 blacklist_dbs=None, max_syncs_per_minute=None,
                 users=None, role_pattern='"%s_ROLE"', dry_run=False):
    """
    Synchronizes all policies on all tables in the catalog.

    Parameters
    ----------
    planner_conn :
        Okera planner connection.

    db_conn :
        Connection to the underlying DB.

    whitelist_dbs: list[str], optional
        If set, only dbs that exist in the white list are synced.

    blacklist_dbs : list[str], optional
        If set, dbs in this list are not synced.

    max_syncs_per_minute : int, optional
        The maximum number of table syncs we will run per minute, rate
        limiting the process. This function will sleep if we are going
        too fast.

    dry_run : bool, optional
        If true, only run read only commands and output DDL but do not
        mutate any state.
    """

    dbs = planner_conn.list_databases()
    eprint('[sync] Syncing %s total dbs...' % len(dbs))
    total = 0
    for db in dbs:
        if whitelist_dbs is not None and db not in whitelist_dbs:
            eprint('Skipping db, did not match whitelist: %s' % db)
            continue
        if blacklist_dbs is not None and db in blacklist_dbs:
            eprint('Skipping db, part of blacklist: %s' % db)
            continue

        eprint('[sync] Syncing tables in db: %s...' % db)

        tbls = planner_conn.list_datasets(db)
        eprint('[sync] ...iterating over %s tbls.' % len(tbls))
        for tbl in tbls:
            md = tbl.metadata
            name = tbl.name
            if TBL_PROP_SYNC_ENABLED not in md or \
                  'true' != md[TBL_PROP_SYNC_ENABLED].lower():
                eprint('......skipping table which does not have sync enabled: %s.%s' %\
                    (db, name))
                continue
            if TBL_PROP_DRIVER not in md or 'snowflake' != md[TBL_PROP_DRIVER].lower():
                eprint('......skipping non-snowflake table: %s.%s' % (db, name))
                continue

            eprint('[sync] ......syncing table: %s.%s' % (db, name))
            sync_policies(planner_conn, db_conn, db, name, db_role, users=users,
                          role_pattern=role_pattern, dry_run=dry_run)
            total += 1

            # Update some stats
            now = datetime.datetime.utcnow()
            now_str = now.strftime('%Y-%m-%d %H:%M')
            if now_str not in stats:
                stats[now_str] = 0
            stats[now_str] += 1
            sync_per_min = stats[now_str]
            if max_syncs_per_minute and sync_per_min >= max_syncs_per_minute:
                # Sleep the rest of the minute
                eprint('...sleeping the rest of the minute: %ssec' % (60 - now.second))
                time.sleep(60 - now.second)
    return total

def sync_catalog_loop(planner_conn, db_conn, db_role, whitelist_dbs=None,
                      blacklist_dbs=None, max_table_syncs_per_minute=None,
                      min_sync_time_secs=None, users=None, role_pattern='"%s_ROLE"',
                      dry_run=False, iters=None):
    """
    Main loop that is expected to run forever (except for testing) to sync the
    entire catalog.
    """
    stats = {}
    n = 0
    while iters is None or n < iters:
        now = datetime.datetime.utcnow()
        start = time.time()
        eprint('[sync] ----------------------------------------------------------')
        eprint('[sync] Starting full catalog sync at %s...' % now)

        synced = sync_catalog(
            planner_conn, db_conn, db_role, stats, whitelist_dbs=whitelist_dbs,
            blacklist_dbs=blacklist_dbs, max_syncs_per_minute=max_table_syncs_per_minute,
            users=users, role_pattern=role_pattern, dry_run=dry_run)
        elapsed_sec = round(time.time() - start)

        eprint('[sync] Full catalog sync complete @ %s. Elapsed time: %s s' % \
            (datetime.datetime.utcnow(), elapsed_sec))
        eprint('[sync] ...tables synced: %s' % synced)
        eprint('Stats:')
        eprint(stats)
        eprint('[sync] ----------------------------------------------------------')

        if min_sync_time_secs and elapsed_sec < min_sync_time_secs:
            wait_sec = min_sync_time_secs - elapsed_sec
            eprint('[sync] Sleeping %s sec before next sync...' % wait_sec)
            time.sleep(wait_sec)
        n += 1




# Copyright 2019 Okera Inc. All Rights Reserved.
#
# Tests for CTE rewrite
#
#
# Note about v2 vs v3. In v2, the user is expected to author against our
# (via presto) catalog. This means the table and column names are what
# we return, not what is in the native (e.g. snowflake) system.
# In V3, the original user query is expected to be what it would author
# against the native system.
#

import pytest
import sqlparse
import unittest

from inspect import currentframe, getframeinfo

from okera._thrift_api import TAuthorizeQueryClient
from okera._thrift_api import TRecordServiceException

from okera.tests import pycerebro_test_common as common
from okera.tests import snowflake_cte_test_cases

SKIP_LEVELS = ["smoke", "dev", "all", "checkin"]

# Alias to shorten test code
PRESTO = TAuthorizeQueryClient.PRESTO
SNOWFLAKE = TAuthorizeQueryClient.SNOWFLAKE

CLIENT_NAMES = {
  PRESTO: 'presto',
  SNOWFLAKE: 'snowflake'
}

# This is created and set with a "standard" config
TEST_USER = 'testuser'
TEST_ROLE = "cte_sf_test_role"

# Format for test cases. It consists of tuples, each tuple represents a test
# query, test user and a list tuple of expected results. The list contains pairs of
# lists of clients to the expected result.
GENERAL_CASES = [
  # Two examples, first is compatible for all clients, second for snowflake only
#  (getframeinfo(currentframe()).lineno,
#      'SELECT 1', None, [(None, 'SELECT 1')]),
#  (getframeinfo(currentframe()).lineno,
#      'SELECT 1', TEST_USER, [([SNOWFLAKE], 'SELECT 1')]),
  (getframeinfo(currentframe()).lineno,
      """
      select STRING || STRING, concat(STRING, STRING)
      FROM DEMO_DB.JDBC_TEST.ALL_TYPES
      """, None, [
      ([SNOWFLAKE],
"""
SELECT concat(STRING, STRING),
       concat(STRING, STRING)
FROM "DEMO_DB"."JDBC_TEST"."ALL_TYPES" "ALL_TYPES"
""")
  ]),

#  (getframeinfo(currentframe()).lineno,
#      """
#      select '1' + 1, '1'::DOUBLE + 1;
#      FROM DEMO_DB.JDBC_TEST.ALL_TYPES
#      """, None, [
#      ([SNOWFLAKE],
#"""
#SELECT concat(STRING, STRING), concat(STRING, STRING)
#FROM demo_db.jdbc_test.all_types
#""")
#  ]),

  # Iterate over all cases, quoting and aliases (cased and quoted)
  (getframeinfo(currentframe()).lineno,
      'SELECT NATIONKEY FROM TPCH_TEST.PUBLIC.NATION', None, [
      ([SNOWFLAKE], 'SELECT NATIONKEY FROM "TPCH_TEST"."PUBLIC"."NATION" "NATION"')
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT nationkey FROM TPCH_TEST.PUBLIC.NATION', None, [
      ([SNOWFLAKE], 'SELECT nationkey FROM "TPCH_TEST"."PUBLIC"."NATION" "NATION"')
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT nationKey FROM TPCH_TEST.PUBLIC.NATION', None, [
      ([SNOWFLAKE], 'SELECT nationKey FROM "TPCH_TEST"."PUBLIC"."NATION" "NATION"')
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT "NATIONKEY" FROM TPCH_TEST.PUBLIC.NATION', None, [
      ([SNOWFLAKE], 'SELECT "NATIONKEY" FROM "TPCH_TEST"."PUBLIC"."NATION" "NATION"')
  ]),
  # This is invalid directly against snowflake due to the original query's "nationKey"
  # (wrong case and quotes)
  (getframeinfo(currentframe()).lineno,
      'SELECT "nationKey" FROM TPCH_TEST.PUBLIC.NATION', None, [
      ([SNOWFLAKE], None)
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT NATIONKEY as N FROM TPCH_TEST.PUBLIC.NATION', None, [
      ([SNOWFLAKE], 'SELECT NATIONKEY N FROM "TPCH_TEST"."PUBLIC"."NATION" "NATION"')
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT NATIONKEY N FROM TPCH_TEST.PUBLIC.NATION', None, [
      ([SNOWFLAKE], 'SELECT NATIONKEY N FROM "TPCH_TEST"."PUBLIC"."NATION" "NATION"')
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT NATIONKEY n FROM TPCH_TEST.PUBLIC.NATION', None, [
      ([SNOWFLAKE], 'SELECT NATIONKEY n FROM "TPCH_TEST"."PUBLIC"."NATION" "NATION"')
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT NATIONKEY "N" FROM TPCH_TEST.PUBLIC.NATION', None, [
      ([SNOWFLAKE], 'SELECT NATIONKEY "N" FROM "TPCH_TEST"."PUBLIC"."NATION" "NATION"')
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT NATIONKEY "n" FROM TPCH_TEST.PUBLIC.NATION', None, [
      ([SNOWFLAKE], 'SELECT NATIONKEY "n" FROM "TPCH_TEST"."PUBLIC"."NATION" "NATION"')
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT nationkey FROM TPCH_TEST.PUBLIC.NATION', None, [
      ([SNOWFLAKE], 'SELECT nationkey FROM "TPCH_TEST"."PUBLIC"."NATION" "NATION"')
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT nationKey FROM TPCH_TEST.PUBLIC.NATION', None, [
      ([SNOWFLAKE], 'SELECT nationKey FROM "TPCH_TEST"."PUBLIC"."NATION" "NATION"')
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT "NATIONKEY" FROM TPCH_TEST.PUBLIC.NATION', None, [
      ([SNOWFLAKE], 'SELECT "NATIONKEY" FROM "TPCH_TEST"."PUBLIC"."NATION" "NATION"')
  ]),
  # This is invalid directly against snowflake due to the original query's, "nationKey"
  # (wrong case and quotes)
  (getframeinfo(currentframe()).lineno,
      'SELECT "nationKey" FROM TPCH_TEST.PUBLIC.NATION', None, [
      ([SNOWFLAKE], None)
  ]),
  # This is invalid directly against snowflake due to the original "nationKey"
  # (wrong case and quotes)
  (getframeinfo(currentframe()).lineno,
      'SELECT "nationKey" "n" FROM TPCH_TEST.PUBLIC.NATION', None, [
      ([SNOWFLAKE], None)
  ]),

  # same as above but  for test user
  (getframeinfo(currentframe()).lineno,
      'SELECT NATIONKEY FROM TPCH_TEST.PUBLIC.NATION', TEST_USER, [
      ([SNOWFLAKE],
"""
WITH okera_rewrite_tpch_test_snowflake__nation AS
  (SELECT okera_udfs.public.tokenize("NATIONKEY", last_query_id()) as "NATIONKEY"
     FROM "TPCH_TEST"."PUBLIC"."NATION")
SELECT NATIONKEY
FROM okera_rewrite_tpch_test_snowflake__nation "NATION"
""")
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT nationkey FROM TPCH_TEST.PUBLIC.NATION', TEST_USER, [
      ([SNOWFLAKE],
"""
WITH okera_rewrite_tpch_test_snowflake__nation AS
  (SELECT okera_udfs.public.tokenize("NATIONKEY", last_query_id()) as "NATIONKEY"
     FROM "TPCH_TEST"."PUBLIC"."NATION")
SELECT nationkey
FROM okera_rewrite_tpch_test_snowflake__nation "NATION"
""")
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT nationKey FROM TPCH_TEST.PUBLIC.NATION', TEST_USER, [
      ([SNOWFLAKE],
"""
WITH okera_rewrite_tpch_test_snowflake__nation AS
  (SELECT okera_udfs.public.tokenize("NATIONKEY", last_query_id()) as "NATIONKEY"
     FROM "TPCH_TEST"."PUBLIC"."NATION")
SELECT nationKey
FROM okera_rewrite_tpch_test_snowflake__nation "NATION"
""")
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT "NATIONKEY" FROM TPCH_TEST.PUBLIC.NATION', TEST_USER, [
      ([SNOWFLAKE],
"""
WITH okera_rewrite_tpch_test_snowflake__nation AS
  (SELECT okera_udfs.public.tokenize("NATIONKEY", last_query_id()) as "NATIONKEY"
     FROM "TPCH_TEST"."PUBLIC"."NATION")
SELECT "NATIONKEY"
FROM okera_rewrite_tpch_test_snowflake__nation "NATION"
""")
  ]),
# This is invalid directly against snowflake due to the original query's "nationKey"
# (wrong case and quotes)
#  (getframeinfo(currentframe()).lineno,
#      'SELECT "nationKey" FROM TPCH_TEST.PUBLIC.NATION', TEST_USER, [
#      ([SNOWFLAKE],
#"""
#WITH okera_rewrite_tpch_test_snowflake__nation AS
#  (SELECT okera_udfs.public.tokenize("NATIONKEY", last_query_id()) as "NATIONKEY"
#     FROM "TPCH_TEST"."PUBLIC"."NATION")
#SELECT "nationKey"
#FROM okera_rewrite_tpch_test_snowflake__nation "NATION"
#""")
#  ]),

  (getframeinfo(currentframe()).lineno,
      'SELECT NATIONKEY as N FROM TPCH_TEST.PUBLIC.NATION', TEST_USER, [
      ([SNOWFLAKE],
"""
WITH okera_rewrite_tpch_test_snowflake__nation AS
  (SELECT okera_udfs.public.tokenize("NATIONKEY", last_query_id()) as "NATIONKEY"
     FROM "TPCH_TEST"."PUBLIC"."NATION")
SELECT NATIONKEY N
FROM okera_rewrite_tpch_test_snowflake__nation "NATION"
""")
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT NATIONKEY N FROM TPCH_TEST.PUBLIC.NATION', TEST_USER, [
      ([SNOWFLAKE],
"""
WITH okera_rewrite_tpch_test_snowflake__nation AS
  (SELECT okera_udfs.public.tokenize("NATIONKEY", last_query_id()) as "NATIONKEY"
     FROM "TPCH_TEST"."PUBLIC"."NATION")
SELECT NATIONKEY N
FROM okera_rewrite_tpch_test_snowflake__nation "NATION"
""")
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT NATIONKEY n FROM TPCH_TEST.PUBLIC.NATION', TEST_USER, [
      ([SNOWFLAKE],
"""
WITH okera_rewrite_tpch_test_snowflake__nation AS
  (SELECT okera_udfs.public.tokenize("NATIONKEY", last_query_id()) as "NATIONKEY"
     FROM "TPCH_TEST"."PUBLIC"."NATION")
SELECT NATIONKEY n
FROM okera_rewrite_tpch_test_snowflake__nation "NATION"
""")
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT NATIONKEY "N" FROM TPCH_TEST.PUBLIC.NATION', TEST_USER, [
      ([SNOWFLAKE],
"""
WITH okera_rewrite_tpch_test_snowflake__nation AS
  (SELECT okera_udfs.public.tokenize("NATIONKEY", last_query_id()) as "NATIONKEY"
     FROM "TPCH_TEST"."PUBLIC"."NATION")
SELECT NATIONKEY "N"
FROM okera_rewrite_tpch_test_snowflake__nation "NATION"
""")
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT NATIONKEY "n" FROM TPCH_TEST.PUBLIC.NATION', TEST_USER, [
      ([SNOWFLAKE],
"""
WITH okera_rewrite_tpch_test_snowflake__nation AS
  (SELECT okera_udfs.public.tokenize("NATIONKEY", last_query_id()) as "NATIONKEY"
     FROM "TPCH_TEST"."PUBLIC"."NATION")
SELECT NATIONKEY "n"
FROM okera_rewrite_tpch_test_snowflake__nation "NATION"
""")
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT nationkey FROM TPCH_TEST.PUBLIC.NATION', TEST_USER, [
      ([SNOWFLAKE],
"""
WITH okera_rewrite_tpch_test_snowflake__nation AS
  (SELECT okera_udfs.public.tokenize("NATIONKEY", last_query_id()) as "NATIONKEY"
     FROM "TPCH_TEST"."PUBLIC"."NATION")
SELECT nationkey
FROM okera_rewrite_tpch_test_snowflake__nation "NATION"
""")
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT nationkey FROM TPCH_TEST.PUBLIC.NATION', TEST_USER, [
      ([SNOWFLAKE],
"""
WITH okera_rewrite_tpch_test_snowflake__nation AS
  (SELECT okera_udfs.public.tokenize("NATIONKEY", last_query_id()) as "NATIONKEY"
     FROM "TPCH_TEST"."PUBLIC"."NATION")
SELECT nationkey
FROM okera_rewrite_tpch_test_snowflake__nation "NATION"
""")
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT "NATIONKEY" FROM TPCH_TEST.PUBLIC.NATION', TEST_USER, [
      ([SNOWFLAKE],
"""
WITH okera_rewrite_tpch_test_snowflake__nation AS
  (SELECT okera_udfs.public.tokenize("NATIONKEY", last_query_id()) as "NATIONKEY"
     FROM "TPCH_TEST"."PUBLIC"."NATION")
SELECT "NATIONKEY"
FROM okera_rewrite_tpch_test_snowflake__nation "NATION"
""")
  ]),
  (getframeinfo(currentframe()).lineno,
      'SELECT "NATIONKEY" "n" FROM TPCH_TEST.PUBLIC.NATION', TEST_USER, [
      ([SNOWFLAKE],
"""
WITH okera_rewrite_tpch_test_snowflake__nation AS
  (SELECT okera_udfs.public.tokenize("NATIONKEY", last_query_id()) as "NATIONKEY"
     FROM "TPCH_TEST"."PUBLIC"."NATION")
SELECT "NATIONKEY" "n"
FROM okera_rewrite_tpch_test_snowflake__nation "NATION"
""")
  ]),
  (getframeinfo(currentframe()).lineno,
      """
      select
          "a11"."SUPPKEY" "SUPPKEY",
          (sum("a11"."AVAILQTY") + sum("a11"."SUPPLYCOST")) "WJXBFS1"
          from    "TPCH_TEST"."PUBLIC"."PARTSUPP"    "a11"
          where   "a11"."SUPPKEY" = 1
          group by    "a11"."SUPPKEY"
      """, TEST_USER, [
      ([SNOWFLAKE],
"""
WITH okera_rewrite_tpch_test_snowflake__partsupp AS
  (SELECT okera_udfs.public.tokenize("SUPPKEY", last_query_id()) as "SUPPKEY",
          "AVAILQTY",
          "SUPPLYCOST"
   FROM "TPCH_TEST"."PUBLIC"."PARTSUPP")
SELECT "a11"."SUPPKEY" "SUPPKEY",
       (sum("a11"."AVAILQTY") + sum("a11"."SUPPLYCOST")) "WJXBFS1"
FROM okera_rewrite_tpch_test_snowflake__partsupp "a11"
WHERE ("a11"."SUPPKEY" = 1)
GROUP BY "a11"."SUPPKEY"
""")
  ]),
]

#
# These cases as nested how we alias the transformed column in the CTE block. We
# need to name these columns, preserving the original identifier so that the
# original user query works. In otherwise, the names of the columns in the CTE need
# to match the original table.
#
# For snowflake: our strategy is to always quote the original table column name
# as the alias. This works because:
#   1. Quoted upper cases is ignored, its the same as non-quoted upper case
#   2. Originally non-quoted columns in SF as stored as upper case.
#
IDENTIFIER_ALIASES = [
  (getframeinfo(currentframe()).lineno,
      'SELECT * FROM DEMO_DB.JDBC_TEST.MIXED_IDENTIFIERS', TEST_USER, [
      ([SNOWFLAKE],
"""
WITH okera_rewrite_jdbc_test_snowflake__mixed_identifiers AS
  (SELECT CAST(0 AS BIGINT) as "LOWER_NO_QUOTES",
          CAST(0 AS BIGINT) as "lower_quotes",
          CAST(0 AS BIGINT) as "MIXEDNOQUOTES",
          CAST(0 AS BIGINT) as "mixedQuotes",
          CAST(0 AS BIGINT) as "UPPER_NO_QUOTES",
          CAST(0 AS BIGINT) as "UPPER_QUOTES"
   FROM "DEMO_DB"."JDBC_TEST"."MIXED_IDENTIFIERS")
SELECT *
FROM okera_rewrite_jdbc_test_snowflake__mixed_identifiers "MIXED_IDENTIFIERS"
""")
  ]),
  (getframeinfo(currentframe()).lineno,
      """SELECT LOWER_NO_QUOTES,
              "lower_quotes",
              mixedNoQuotes,
              "mixedQuotes",
              UPPER_NO_QUOTES,
              "UPPER_QUOTES"
              FROM DEMO_DB.JDBC_TEST.MIXED_IDENTIFIERS""", TEST_USER, [
      ([SNOWFLAKE],
"""
WITH okera_rewrite_jdbc_test_snowflake__mixed_identifiers AS
  (SELECT CAST(0 AS BIGINT) as "LOWER_NO_QUOTES",
          CAST(0 AS BIGINT) as "lower_quotes",
          CAST(0 AS BIGINT) as "MIXEDNOQUOTES",
          CAST(0 AS BIGINT) as "mixedQuotes",
          CAST(0 AS BIGINT) as "UPPER_NO_QUOTES",
          CAST(0 AS BIGINT) as "UPPER_QUOTES"
   FROM "DEMO_DB"."JDBC_TEST"."MIXED_IDENTIFIERS")
SELECT LOWER_NO_QUOTES,
      "lower_quotes",
      mixedNoQuotes,
      "mixedQuotes",
      UPPER_NO_QUOTES,
      "UPPER_QUOTES"
FROM okera_rewrite_jdbc_test_snowflake__mixed_identifiers "MIXED_IDENTIFIERS"
""")
  ]),

# For columns not originally case-sensitive, try some permutations. The case sensitive
# ones fail in snowflake originally, as expected, if they don't match.
  (getframeinfo(currentframe()).lineno,
      """SELECT LOWER_NO_QUOTES,
              lower_no_QUOTES,
              mixedNoQuotes,
              mixednoquotes,
              UPPER_NO_QUOTES,
              upper_NO_quotes
              FROM DEMO_DB.JDBC_TEST.MIXED_IDENTIFIERS""", TEST_USER, [
      ([SNOWFLAKE],
"""
WITH okera_rewrite_jdbc_test_snowflake__mixed_identifiers AS
  (SELECT CAST(0 AS BIGINT) as "LOWER_NO_QUOTES",
          CAST(0 AS BIGINT) as "MIXEDNOQUOTES",
          CAST(0 AS BIGINT) as "UPPER_NO_QUOTES"
   FROM "DEMO_DB"."JDBC_TEST"."MIXED_IDENTIFIERS")
SELECT LOWER_NO_QUOTES,
       lower_no_QUOTES,
       mixedNoQuotes,
       mixednoquotes,
       UPPER_NO_QUOTES,
       upper_NO_quotes
FROM okera_rewrite_jdbc_test_snowflake__mixed_identifiers "MIXED_IDENTIFIERS"
""")
  ]),
]

class AuthorizeQueryTest(common.TestBase):
    def role_setup(self, conn):
      self._recreate_test_role(conn, TEST_ROLE, [TEST_USER])
      ddls = [
        'DROP ATTRIBUTE IF EXISTS snowflake_test.attr',
        'CREATE ATTRIBUTE snowflake_test.attr',
        'ALTER TABLE tpch_test_snowflake.nation ADD COLUMN ATTRIBUTE nationkey %s'\
             % 'snowflake_test.attr',
        'ALTER TABLE tpch_test_snowflake.partsupp ADD COLUMN ATTRIBUTE suppkey %s'\
             % 'snowflake_test.attr',
        'ALTER TABLE jdbc_test_snowflake.mixed_identifiers ADD ATTRIBUTE %s'\
             % 'snowflake_test.attr',
        ('GRANT SELECT ON TABLE tpch_test_snowflake.nation TRANSFORM %s ' +
         'WITH %s TO ROLE %s') % ('snowflake_test.attr', 'tokenize()', TEST_ROLE),
        ('GRANT SELECT ON TABLE tpch_test_snowflake.partsupp TRANSFORM %s ' +
         'WITH %s TO ROLE %s') % ('snowflake_test.attr', 'tokenize()', TEST_ROLE),
        ('GRANT SELECT ON TABLE jdbc_test_snowflake.mixed_identifiers TRANSFORM %s ' +
         'WITH %s TO ROLE %s') % ('snowflake_test.attr', 'zero()', TEST_ROLE),
      ]
      for ddl in ddls:
          conn.execute_ddl(ddl)

    def cte(self, conn, sql, client, user=None):
      result, plan, _, _ = self.cte_rewrite(
          conn, sql, user=user, client=client, generate_plan=True,
          validate_rewrite=True)
      self.assertTrue(plan is not None)
      return result

    # Runs all queries in cases that are compatible with target_client
    def run_client(self, cases, conn, target_client, user=None, test_case=None):
        for (lineno, sql, u, results) in cases:
            if u != user:
                continue
            if test_case != None and test_case != lineno:
                continue

            for clients, expected in results:
                if clients == None or target_client in clients:
                    self.role_setup(conn)

                    header = "\n------------ Test Case@%s (%s, %s) ----------------\n" % \
                        (lineno, user, CLIENT_NAMES[target_client])
                    header += sql + '\n'
                    print(header)
                    if expected:
                        result = self.cte(conn, sql, user=user, client=target_client)
                        self.assert_sql_equals(expected, result, error_header=header)
                    else:
                        # We currently can't run these tests. These require turning off ignoring
                        # case sensitivity in the driver, which breaks other tests.
                        return
                        #with self.assertRaises(TRecordServiceException) as ex:
                        #    result = self.cte(conn, sql, user=user, client=target_client)
                        #    header += 'Expected query to fail but did not.\nResult:\n'
                        #    header += result
                        #    self.assertFalse(True, msg=header)

    @unittest.skip("Ad-hoc harness for development to try a query, not a test")
    def test_snowflake_adhoc(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            sql = """
            select
                "a11"."SUPPKEY" "SUPPKEY",
                (sum("a11"."AVAILQTY") + sum("a11"."SUPPLYCOST")) "WJXBFS1"
                from    "TPCH_TEST"."PUBLIC"."PARTSUPP"    "a11"
                where   "a11"."SUPPKEY" = 1
                group by    "a11"."SUPPKEY"
            """
            print()
            print(sqlparse.format(self.cte(conn, sql, SNOWFLAKE), reindent=True))

    @unittest.skip("Does not work in CI, multi connections, same table.")
    def test_snowflake_general(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self.run_client(GENERAL_CASES, conn, SNOWFLAKE)
            self.run_client(GENERAL_CASES, conn, SNOWFLAKE, TEST_USER)

    @unittest.skipIf(common.should_skip(SKIP_LEVELS, "snowflake-cte"), "Skip at level")
    def test_snowflake_v2(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            for (f, line, sql, expected) in snowflake_cte_test_cases.V2_PRESTO_TESTS:
                f = f.split('/')[-1]
                header = "\n--------------- Test Case@%s:%s -------------------\n" % \
                    (f, line)
                header += sql + '\n'
                print(header)
                result = self.cte(conn, sql, client=PRESTO)
                self.assert_sql_equals(expected, result, error_header=header)

    @unittest.skipIf(common.should_skip(SKIP_LEVELS, "snowflake-cte"), "Skip at level")
    def test_snowflake_identifiers(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self.run_client(IDENTIFIER_ALIASES, conn, SNOWFLAKE, TEST_USER)

    @unittest.skipIf(common.should_skip(SKIP_LEVELS, "snowflake-cte"), "Skip at level")
    def test_snowflake_policy_v2(self):
        self.maxDiff = None
        ctx = common.get_test_context()

        # Tuple of (policy, query, expected rewrite)
        policies = [
            # Try a few cases and quotes on the STRING column
            ("""
              GRANT SELECT ON TABLE jdbc_test_snowflake.all_types
              WHERE if (`int` > 100, true, false)
              TO ROLE sf_test_role
            """,
            """SELECT "STRING" FROM jdbc_test_snowflake.all_types""",
            """
            WITH okera_rewrite_jdbc_test_snowflake__all_types AS
                (SELECT "STRING" as "string"
                FROM "DEMO_DB"."JDBC_TEST"."ALL_TYPES"
                WHERE iff("INT" > 100, TRUE, FALSE))
            SELECT "STRING" FROM okera_rewrite_jdbc_test_snowflake__all_types "all_types"
            """),
            ("""
              GRANT SELECT ON TABLE jdbc_test_snowflake.all_types
              WHERE if (`int` > 100, true, false)
              TO ROLE sf_test_role
            """,
            """SELECT STRING FROM jdbc_test_snowflake.all_types""",
            """
            WITH okera_rewrite_jdbc_test_snowflake__all_types AS
                (SELECT "STRING" as "string"
                FROM "DEMO_DB"."JDBC_TEST"."ALL_TYPES"
                WHERE iff("INT" > 100, TRUE, FALSE))
            SELECT STRING FROM okera_rewrite_jdbc_test_snowflake__all_types "all_types"
            """),
            ("""
              GRANT SELECT ON TABLE jdbc_test_snowflake.all_types
              WHERE if (`int` > 100, true, false)
              TO ROLE sf_test_role
            """,
            """SELECT "string" FROM jdbc_test_snowflake.all_types""",
            """
            WITH okera_rewrite_jdbc_test_snowflake__all_types AS
                (SELECT "STRING" as "string"
                FROM "DEMO_DB"."JDBC_TEST"."ALL_TYPES"
                WHERE iff("INT" > 100, TRUE, FALSE))
            SELECT "string" FROM okera_rewrite_jdbc_test_snowflake__all_types "all_types"
            """),

            ("""
              GRANT SELECT ON TABLE jdbc_test_snowflake.all_types
              WHERE if (`int` > 100, true, false)
              TO ROLE sf_test_role
            """,
            """SELECT "STRING" FROM jdbc_test_snowflake.all_types
               WHERE IF ("INT" > 1, true, false)""",
            """
            WITH okera_rewrite_jdbc_test_snowflake__all_types AS
              (SELECT "STRING" as "string", "INT" as "int"
               FROM "DEMO_DB"."JDBC_TEST"."ALL_TYPES"
               WHERE iff("INT" > 100, TRUE, FALSE))
            SELECT "STRING"
            FROM okera_rewrite_jdbc_test_snowflake__all_types "all_types"
            WHERE IFF(("INT" > 1), true, false)
            """),

            ("""
              GRANT SELECT ON TABLE jdbc_test_snowflake.all_types
              WHERE sets_intersect(`string`, 'hello') OR
                    sets_intersect(`string`, user_attribute('not_exist'))
              TO ROLE sf_test_role""",
            """SELECT "STRING" FROM jdbc_test_snowflake.all_types""",
            """
              WITH okera_rewrite_jdbc_test_snowflake__all_types AS
                (SELECT "STRING" as "string" FROM
                "DEMO_DB"."JDBC_TEST"."ALL_TYPES"
                WHERE arrays_overlap(split("STRING", ','), split('hello', ','))
                OR arrays_overlap(split("STRING", ','), split(NULL, ',')))
              SELECT "STRING" FROM
              okera_rewrite_jdbc_test_snowflake__all_types "all_types"
            """),
            ("""
              GRANT SELECT ON TABLE jdbc_test_snowflake.array_test
              WHERE array_contains(to_variant('donald'), to_array(column1))
              TO ROLE sf_test_role""",
            """SELECT * from jdbc_test_snowflake.array_test""",
            """
              WITH okera_rewrite_jdbc_test_snowflake__array_test AS (SELECT
              "ID" as "id",
              "COLUMN1" as "column1",
              "COLUMN2" as "column2"
              FROM "DEMO_DB"."JDBC_TEST"."ARRAY_TEST"
              WHERE "array_contains"("to_variant"('donald'), "to_array"("COLUMN1")))
              SELECT *
              FROM
              okera_rewrite_jdbc_test_snowflake__array_test "array_test"
            """),
        ]

        with common.get_planner(ctx) as conn:
            conn.execute_ddl("DROP ATTRIBUTE IF EXISTS snowflake_test.attr")
            conn.execute_ddl("CREATE ATTRIBUTE snowflake_test.attr")
            for policy, query, expected in policies:
                print("\n------------------ Test Case -------------------")
                print("Policy:")
                print(policy)
                print("Query:")
                print(query)
                self._recreate_test_role(conn, 'sf_test_role', ['sf_testuser'])
                conn.execute_ddl(policy)

                result = self.cte(conn, query,
                    user='sf_testuser', client=TAuthorizeQueryClient.PRESTO)
                self.assert_sql_equals(expected, result)

    @unittest.skipIf(common.should_skip(SKIP_LEVELS, "snowflake-cte"), "Skip at level")
    def test_snowflake_policy_v3(self):
        self.maxDiff = None
        ctx = common.get_test_context()

        # Tuple of (policy, query, expected rewrite)
        policies = [
            ("""
              GRANT SELECT ON TABLE jdbc_test_snowflake.all_types
              WHERE if (`int` > 100, true, false)
              TO ROLE sf_test_role
            """,
            """SELECT "STRING" FROM DEMO_DB.JDBC_TEST.ALL_TYPES""",
            """
            WITH okera_rewrite_jdbc_test_snowflake__all_types AS
                (SELECT "STRING" FROM "DEMO_DB"."JDBC_TEST"."ALL_TYPES"
                WHERE iff("INT" > 100, TRUE, FALSE))
            SELECT "STRING" FROM okera_rewrite_jdbc_test_snowflake__all_types "ALL_TYPES"
            """),

            ("""
              GRANT SELECT ON TABLE jdbc_test_snowflake.all_types
              WHERE if (`int` > 100, true, false)
              TO ROLE sf_test_role
            """,
            """SELECT "STRING" FROM "DEMO_DB"."JDBC_TEST"."ALL_TYPES"
               WHERE IF ("INT" > 1, true, false)""",
            """
            WITH okera_rewrite_jdbc_test_snowflake__all_types AS
              (SELECT "STRING", "INT"
               FROM "DEMO_DB"."JDBC_TEST"."ALL_TYPES"
               WHERE iff("INT" > 100, TRUE, FALSE))
            SELECT "STRING"
            FROM okera_rewrite_jdbc_test_snowflake__all_types "ALL_TYPES"
            WHERE IFF(("INT" > 1), true, false)
            """),

            ("""
              GRANT SELECT ON TABLE jdbc_test_snowflake.all_types
              WHERE sets_intersect(`string`, 'hello') OR
                    sets_intersect(`string`, user_attribute('not_exist'))
              TO ROLE sf_test_role""",
            """SELECT STRING FROM DEMO_DB.JDBC_TEST.ALL_TYPES""",
            """
              WITH okera_rewrite_jdbc_test_snowflake__all_types AS
                (SELECT "STRING" FROM
                "DEMO_DB"."JDBC_TEST"."ALL_TYPES"
                WHERE arrays_overlap(split("STRING", ','), split('hello', ','))
                OR arrays_overlap(split("STRING", ','), split(NULL, ',')))
              SELECT STRING FROM
              okera_rewrite_jdbc_test_snowflake__all_types "ALL_TYPES"
            """),
            ("""
              GRANT SELECT ON TABLE jdbc_test_snowflake.array_test
              WHERE array_contains(to_variant('donald'), to_array(column1))
              TO ROLE sf_test_role""",
            """SELECT * from DEMO_DB.JDBC_TEST.ARRAY_TEST""",
            """
              WITH okera_rewrite_jdbc_test_snowflake__array_test AS (SELECT
                "ID",
                "COLUMN1",
                "COLUMN2"
                FROM "DEMO_DB"."JDBC_TEST"."ARRAY_TEST"
                WHERE "array_contains"("to_variant"('donald'), "to_array"("COLUMN1")))
              SELECT * FROM
              okera_rewrite_jdbc_test_snowflake__array_test "ARRAY_TEST"
            """),
        ]

        with common.get_planner(ctx) as conn:
            conn.execute_ddl("DROP ATTRIBUTE IF EXISTS snowflake_test.attr")
            conn.execute_ddl("CREATE ATTRIBUTE snowflake_test.attr")
            for policy, query, expected in policies:
                print("\n------------------ Test Case -------------------")
                print("Policy:")
                print(policy)
                print("Query:")
                print(query)
                self._recreate_test_role(conn, 'sf_test_role', ['sf_testuser'])
                conn.execute_ddl(policy)

                result = self.cte(conn, query,
                    user='sf_testuser', client=TAuthorizeQueryClient.SNOWFLAKE)
                self.assert_sql_equals(expected, result)

if __name__ == "__main__":
    unittest.main()




# pylint: disable=bare-except
# pylint: disable=consider-using-enumerate
# pylint: disable=len-as-condition
# pylint: disable=protected-access
# pylint: disable=too-few-public-methods
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-locals
# pylint: disable=unidiomatic-typecheck
# pylint: disable=bare-except

import base64
import datetime
import json
import math
import os
import random
import re
import sqlparse
import statistics
import string
import time
import uuid
import unittest

import pg8000
from prettytable import PrettyTable
import pytest
import pytz
import sqlparse

from okera import context, _thrift_api
from okera._thrift_api import TAttribute
from okera._thrift_api import TAuthorizeQueryClient
from okera._thrift_api import TAuthorizeQueryParams
from okera._thrift_api import TDataRegConnection
from okera._thrift_api import TRequestContext
from okera._thrift_api import TTypeId, TConfigType

DEFAULT_PRESTO_HOST = os.environ['ODAS_TEST_HOST']
DEFAULT_PRESTO_PORT = os.environ['ODAS_TEST_PORT_PRESTO_COORD_HTTPS']
DEFAULT_REST_SERVER_HOST = os.environ['ODAS_TEST_HOST']
DEFAULT_REST_SERVER_PORT = os.environ['ODAS_TEST_PORT_REST_SERVER']
DEFAULT_ACCESS_PROXY_HOST = os.environ['ODAS_TEST_HOST']
DEFAULT_ACCESS_PROXY_PORT = os.environ['ODAS_TEST_PORT_ACCESS_PROXY']

PERF_DB_HOST = 'okera-psql-aurora.cluster-cyn8yfvyuugz.us-west-2.rds.amazonaws.com'
PERF_DB_USER = 'okera_perf_user'
PERF_DB_DB = 'rpc_perf_db'

TEST_LEVELS = {
    'notests' : 0,
    'smoke' : 1,
    'dev' : 2,
    'checkin' : 3,
    'ci' : 4,
    'all' : 5,
}

TEST_LEVEL = 'smoke'
if 'TEST_LEVEL' in os.environ:
    TEST_LEVEL = os.environ['TEST_LEVEL']
elif 'CEREBRO_TEST_LEVEL' in os.environ:
    TEST_LEVEL = os.environ['CEREBRO_TEST_LEVEL']
ROOT_TOKEN = os.environ['OKERA_HOME'] + '/integration/tokens/cerebro.token'

OS_VERSION = None
try:
  import lsb_release
  OS_VERSION = lsb_release.get_os_release()['RELEASE']
except:
  pass

identity = lambda x: x

def get_env_var(name, coercer, default):
    if name in os.environ:
        return coercer(os.environ[name])
    return default

def get_bool_env_var(name, default):
    return get_env_var(name, lambda x: str(x).lower() in ['true'], default)

# If true, run the perf benchmarks.
RUN_PERF = get_env_var('RUN_PERF', bool, False)
SKIP_PERF_DATA_LOAD = get_env_var('SKIP_PERF_DATA_LOAD', bool, False)
DEFAULT_ITERS = get_env_var('PERF_ITERATIONS', int, 50)
QUIET = get_env_var("PERF_QUIET", bool, False)
# If true, run the snowflake proxy perf as well
RUN_PERF_SNOWFLAKE = get_env_var("RUN_PERF_SNOWFLAKE", bool, False)

def test_level_lt(min_level):
    if TEST_LEVEL not in TEST_LEVELS:
        return False
    return TEST_LEVELS[TEST_LEVEL] < TEST_LEVELS[min_level]


# List of all known CI test groups (see get_ci_test_cases()). Used to
# validate input.
ALL_ADDITIONAL_CI_TEST_GROUPS = ['redshift-cte', 'snowflake-cte']
def get_ci_test_cases(msg):
    # This parses the gerrit commit message so that additional tests can be run
    # that are not part of standard precommit. This is useful, for example, when
    # changing something in the code that needs additional coverage that is normally
    # run at higher CI.
    # Input should be base64 encoded message (that gerrit produces normally)
    # This parses out for CI_TEST_CASES: [comma list of cases]
    if not msg:
        return []
    m = base64.b64decode(msg).decode('ascii')
    result = []
    for l in m.split('\n'):
        if l.startswith('ADDITIONAL_CI_TESTS:'):
            cases = l.split('ADDITIONAL_CI_TESTS:')[1]
            for c in cases.split(','):
                if c:
                    c = c.strip()
                    if c not in ALL_ADDITIONAL_CI_TEST_GROUPS:
                        raise Exception("Invalid CI test group.")
                    result.append(c)
    return result

ADDITIONAL_CI_TEST_GROUPS = get_ci_test_cases(
    get_env_var('GERRIT_CHANGE_COMMIT_MESSAGE', identity, None))
print('Additional CI Tests: %s' % ADDITIONAL_CI_TEST_GROUPS)

def should_skip(skip_levels, test_group):
    if test_group not in ALL_ADDITIONAL_CI_TEST_GROUPS:
        raise Exception("Invalid CI test group.")
    if test_group in ADDITIONAL_CI_TEST_GROUPS:
        return False
    return TEST_LEVEL in skip_levels

def get_test_context(auth_mech=None, dialect=None, namespace='okera', tz=pytz.utc):
    if auth_mech is None:
        auth_mech = get_env_var('PYCEREBRO_TEST_AUTH_MECH', identity, 'NOSASL')

    ctx = context(dialect=dialect, namespace=namespace, tz=tz)
    if auth_mech == 'NOSASL':
        ctx.disable_auth()
    elif auth_mech == 'TOKEN':
        ctx.enable_token_auth(token_file=ROOT_TOKEN)
    else:
        ctx.disable_auth()
    return ctx

def get_planner(ctx, host=None, port=None, dialect='okera', presto_port=None,
                namespace=None):
    if host is not None:
        host = host
    else:
        host = get_env_var('ODAS_TEST_HOST', identity, 'localhost')

    if port is not None:
        port = port
    else:
        port = get_env_var('ODAS_TEST_PORT_PLANNER_THRIFT', int, 12050)
    if 'presto' in dialect:
        ctx.enable_token_auth(token_str='root')
        return ctx.connect(host=host,
                           port=port,
                           presto_host=host,
                           presto_port=DEFAULT_PRESTO_PORT,
                           namespace=namespace)
    return ctx.connect(host=host, port=port, presto_port=presto_port, namespace=namespace)

def get_worker(ctx, host=None, port=None):
    if host is not None:
        host = host
    else:
        host = get_env_var('ODAS_TEST_HOST', identity, 'localhost')

    if port is not None:
        port = port
    else:
        port = get_env_var('ODAS_TEST_PORT_WORKER_THRIFT', int, 13050)

    return ctx.connect_worker(host=host, port=port)

def get_rest_server_url(scheme='http',
                        host=DEFAULT_REST_SERVER_HOST, port=DEFAULT_REST_SERVER_PORT):
    return "%s://%s:%s" % (scheme, host, port)

def get_access_proxy_url(scheme='http',
                         host=DEFAULT_ACCESS_PROXY_HOST, port=DEFAULT_ACCESS_PROXY_PORT):
    return "%s://%s:%s" % (scheme, host, port)

def measure_latency(iters, fn, msg=None, quiet=False):
    durations = []
    for _ in range(0, iters):
        start = time.time()
        fn()
        durations.append((time.time() - start) * 1000)
    durations = sorted(durations)
    if quiet or QUIET:
        return durations
    if msg:
      print("\n%s\n    Iterations: %s" % (msg, iters))
    else:
      print("\nIterations " + str(iters))
    print("Mean " + str(statistics.mean(durations)) + " ms")
    print("50%: " + str(durations[int(len(durations) * .5)]) + " ms")
    print("90%: " + str(durations[int(len(durations) * .90)]) + " ms")
    print("95%: " + str(durations[int(len(durations) * .95)]) + " ms")
    print("99%: " + str(durations[int(len(durations) * .99)]) + " ms")
    print("99.5%: " + str(durations[int(len(durations) * .995)]) + " ms")
    print("99.9%: " + str(durations[int(len(durations) * .999)]) + " ms")
    return durations

def configure_botocore_patch():
    os.environ['OKERA_PATCH_BOTO'] = 'True'
    os.environ['OKERA_PLANNER_HOST'] = \
        get_env_var('ODAS_TEST_HOST', identity, 'localhost')
    os.environ['OKERA_PLANNER_PORT'] = \
        get_env_var('ODAS_TEST_PORT_PLANNER_THRIFT', identity, 12050)
    from okera import initialize_default_context, check_and_patch_botocore
    initialize_default_context()
    check_and_patch_botocore()

def upsert_config(conn, config_type, key, value):
    """Upsert a configuration.

    config_type : TConfigType
        The type of configurations to return.
    key : str
        The key for this configuration. This must be unique with config_type.
    value : str
        The value for this config.

    Returns
    -------
    bool
        Returns true if the config was updated or false if it was inserted.
    list(str), optional
        List of warnings that were generated as part of this request.
    """
    request = _thrift_api.TConfigUpsertParams()
    request.config_type = config_type
    request.key = key
    if isinstance(value, str):
        request.params = {'value': value}
    else:
        request.params = value
    result = conn.service.client.UpsertConfig(request)
    # TODO: server needs to return if it was an upsert or not
    return True, result.warnings

def delete_config(conn, config_type, key):
    """Upsert a configuration.

    config_type : TConfigType
        The type of configurations to return.
    key : str
        The key for this configuration. This must be unique with config_type.

    Returns
    -------
    bool
        Returns true if the config was deleted.
    list(str), optional
        List of warnings that were generated as part of this request.
    """
    request = _thrift_api.TConfigDeleteParams()
    request.config_type = config_type
    request.key = key
    result = conn.service.client.DeleteConfig(request)
    # TODO: server needs to return if it was deleted or not
    return True, result.warnings

def list_configs(conn, config_type):
    """List configurations of the specified type.

    config_type : TConfigType
        The type of configurations to return.

    Returns
    -------
    map(str, str)
        List of configs as a map of key values.
    """
    table_name = None
    if config_type == TConfigType.AUTOTAGGER_REGEX:
        table_name = "okera_system.tagging_rules"
    elif config_type == TConfigType.SYSTEM_CONFIG:
        table_name = "okera_system.configs"
    else:
        raise ValueError("Invalid config type.")
    return conn.scan_as_json(table_name)

def read_sql_file(path, **kwargs):
    sql = open(path, 'r').read()
    return sql.format(**kwargs)

class PerfResults():
    """
    Utility class for running, storing  and reporing perf results
    """
    def __init__(self):
        self.results = PrettyTable(
          ['rpc', 'name', 'user', 'mean', '50%', '90%', '99%', 'max', 'iters']
        )

    def measure(self, fn, msg, user, rpc, iters=DEFAULT_ITERS):
        """
        Measure and store the latency of fn().
          msg: Name  of test cases
          user: User (typically admin or not, running the function)
          rpc: Name of RPC/API being used.
        """
        if not RUN_PERF:
            return
        label = "%s(%s)" % (msg, user)
        latencies = measure_latency(iters, fn, msg=label)
        mean = "{:.3f}".format(statistics.mean(latencies))
        p50 = "{:.3f}".format(latencies[int(len(latencies) * .5)])
        p90 = "{:.3f}".format(latencies[int(len(latencies) * .9)])
        p99 = "{:.3f}".format(latencies[int(len(latencies) * .99)])
        m = "{:.3f}".format(latencies[-1])
        row = [rpc, msg, user, mean, p50, p90, p99, m, iters]
        self.results.add_row(row)

    def finalize_results(self):
        if not  RUN_PERF:
            return
        print("\nPerf tests done:")
        print(self.results)

        # We want some globals that are the same across python modules so we can
        # correlate the entire run. This can be done by storing a global inside
        # the pytest object.
        globals = {}
        if hasattr(pytest, 'okera_globals'):
            globals = pytest.okera_globals
        else:
            # First time, generate it
            globals['now'] = datetime.datetime.now()
            globals['id'] = uuid.uuid1()
            pytest.okera_globals = globals

        if 'CEREBRO_PERF_DB_PASSWORD' in os.environ:
            self.persist_results(globals)

    def skip_data_load(self):
        return SKIP_PERF_DATA_LOAD

    def run_perf(self):
        return RUN_PERF

    def run_perf_snowflake(self):
        return RUN_PERF and RUN_PERF_SNOWFLAKE

    def persist_results(self, globals):
        print("Persisting results to database.")
        id = globals['id']
        now = globals['now']
        build = -1
        version = 'latest'
        owner = ''
        if 'BUILD_NUMBER' in os.environ:
            build = int(os.environ['BUILD_NUMBER'])
        if 'ODAS_DOCKER_VERSION' in os.environ:
            v = os.environ['ODAS_DOCKER_VERSION']
            if 'jenkins' not in v:
                # Jenkins generates latest versions that are unique, ignore
                version = v
        if 'OKERA_BENCHMARK_OWNER' in os.environ:
            owner = os.environ['OKERA_BENCHMARK_OWNER']
        conn = pg8000.connect(host=PERF_DB_HOST,
                            user=PERF_DB_USER,
                            password=os.environ['CEREBRO_PERF_DB_PASSWORD'],
                            database=PERF_DB_DB,
                            ssl=True)
        cur = conn.cursor()
        for row in self.results:
            row.border = False
            row.header = False
            rpc = row.get_string(fields=['rpc']).strip()
            name = row.get_string(fields=['name']).strip()
            user = row.get_string(fields=['user']).strip()
            mean = row.get_string(fields=['mean']).strip()
            p50 = row.get_string(fields=['50%']).strip()
            p90 = row.get_string(fields=['90%']).strip()
            p99 = row.get_string(fields=['99%']).strip()
            max = row.get_string(fields=['max']).strip()
            iters = row.get_string(fields=['iters']).strip()

            values = (id, now, build, version, owner, rpc,
                      name, user, mean, p50, p90, p99, max, iters)
            sql = """INSERT INTO perf_results
                   VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)"""
            cur.execute(sql, values)
        conn.commit()

class SchemaNode():
  def __init__(self):
      self.col = None
      self.children = []

  @staticmethod
  def convert(cols, idx, schema):
        node = SchemaNode()
        node.col = cols[idx]
        schema.append(node)
        idx += 1
        if node.col.type.num_children:
            for _ in range(0, node.col.type.num_children):
                idx = SchemaNode.convert(cols, idx, node.children)
        return idx

class TmpView():
    def __init__(self, conn, sql):
        self.conn = conn
        # TODO: make this more unique so can be used concurrently
        self.db = "test_tmp_db"
        self.view = "tmp_view"
        self.conn.execute_ddl("CREATE DATABASE IF NOT EXISTS %s" % self.db)
        self.conn.execute_ddl("DROP VIEW IF EXISTS %s.%s" % (self.db, self.view))
        self.conn.execute_ddl("CREATE VIEW %s.%s AS %s" % (self.db, self.view, sql))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.conn.execute_ddl("DROP VIEW IF EXISTS %s.%s" % (self.db, self.view))

    def name(self):
        return '%s.%s' % (self.db, self.view)

class JsonGeneratorNode():
    def __init__(self):
        self.types = []
        self.children = {}

    def to_json(self):
        result = {}
        result['types'] = []
        for t in self.types:
            result['types'].append(TTypeId._VALUES_TO_NAMES[t])
        result['types'] = ', '.join(result['types'])
        if not self.children:
            return result
        for name, child in self.children.items():
            result[name] = child.to_json()
        return result

class JsonGenerator():
    def __init__(self,
                 types=None,
                 record_probabilities=None,
                 array_probabilties=None,
                 null_probability=.1,
                 empty_record_probability=.1,
                 min_fields=1,
                 max_fields=5,
                 max_array_len=3,
                 min_records=2,
                 max_records=10,
                 max_string_len=20,
                 seed=0,
                 max_recursion=5,
                 missing_fields_probability=0,
                 generate_variadic_schema=False,
                 generate_empty_record_all_types=False):
        random.seed(seed)
        if not types:
            types = [[TTypeId.BOOLEAN],
                     [TTypeId.BIGINT],
                     [TTypeId.DOUBLE],
                     [TTypeId.DATE],
                     [TTypeId.TIMESTAMP_NANOS],
                     [TTypeId.STRING]]
            if generate_variadic_schema:
                # This indicates that if this "type" is selected in the schema, the
                # data will be one of these types
                types.append([TTypeId.BIGINT, TTypeId.DOUBLE])
                types.append([TTypeId.BIGINT, TTypeId.DATE])
                types.append([TTypeId.DOUBLE, TTypeId.BOOLEAN])
                types.append([TTypeId.DATE, TTypeId.TIMESTAMP_NANOS])
                types.append([TTypeId.BOOLEAN, TTypeId.BIGINT, TTypeId.DATE,
                              TTypeId.TIMESTAMP_NANOS, TTypeId.DOUBLE, TTypeId.STRING])

        # Probabilities of generating record/array schemas by level. We want to
        # generate them with higher probabilities at the beginning to minimize
        # generating a lot of simple schemas.
        # This generate a CDF (i.e the remaining percentage is used to generate a
        # simple type).
        if not record_probabilities:
            record_probabilities = [.5, .4, .3, .25, .25, .2, .1]
        if not array_probabilties:
            array_probabilities = [.3, .3, .3, .25, .25, .2, .1]

        self.__min_fields = min_fields
        self.__max_fields = max_fields
        self.__min_records = min_records
        self.__max_records = max_records
        self.__max_array_len = max_array_len
        self.__null_probability = null_probability
        self.__max_string_len = max_string_len
        self.__empty_record_probability = empty_record_probability
        self.__max_recursion = max_recursion
        self.__types = types
        self.__array_probabilities = array_probabilities
        self.__record_probabilities = record_probabilities

        # Configs to control schema merge cases
        self.__generate_invalid_schema_merges = False
        self.__generate_empty_record_all_types = generate_empty_record_all_types
        self.__missing_fields_probability = missing_fields_probability

        self.__field_idx = 0
        self.__schema = None

        print("Generating with configuration")
        print("    Seed:  %s" % seed)
        print("    Types:  %s" % self.__types)
        print("    Record Probabilities:  %s" % self.__record_probabilities)
        print("    Array Probabilities:  %s" % self.__array_probabilities)
        print("    Max Array Len: %d" % (self.__max_array_len))
        print("    Max String Len: %d" % (self.__max_string_len))
        print("    Null Probability: %s" % (self.__null_probability))
        print("    Empty Record Probability: %s" % (self.__empty_record_probability))
        print("    Missing Field Probability: %s" % (self.__missing_fields_probability))
        print("    Max Depth: %d" % (self.__max_recursion))
        print("    # Fields: [%d, %d]" % (self.__min_fields, self.__max_fields))
        print("    # Records: [%d, %d]" % (self.__min_records, self.__max_records))
        print("    Generate variadic schemas: %s" % generate_variadic_schema)
        print("    Generate invalid merges: %s" %\
            self.__generate_invalid_schema_merges)
        print("    Generate empty records all types: %s" %\
            self.__generate_empty_record_all_types)

    def new_schema(self):
        self.__field_idx = 0
        self.__schema = self._generate_schema([TTypeId.RECORD], 0)

    def _random_type(self, level):
        prob_record = self.__record_probabilities[\
            min(level, len(self.__record_probabilities) - 1)]
        prob_array = self.__array_probabilities[\
            min(level, len(self.__array_probabilities) - 1)]
        r = random.random()
        if r < prob_record:
            return [TTypeId.RECORD]
        if r < prob_record + prob_array:
            return [TTypeId.ARRAY]
        return random.choice(self.__types)

    def _generate_schema(self, types, level):
        # Recursively generates a test schema
        node = JsonGeneratorNode()
        if level == self.__max_recursion:
            node.types = [TTypeId.STRING]
            return node

        node.types = types
        if types == [TTypeId.RECORD]:
            num_fields = random.randint(self.__min_fields, self.__max_fields - 1)
            for idx in range(0, num_fields):
                t = self._random_type(level)
                if level == 0:
                    name = 'c' + str(idx)
                else:
                    name = 'f' + str(self.__field_idx)
                    self.__field_idx += 1
                node.children[name] = self._generate_schema(t, level + 1)
        elif types == [TTypeId.ARRAY]:
            t = self._random_type(level)
            node.children['item'] = self._generate_schema(t, level + 1)
        return node

    def __generate_random_data(self, schema):
        if random.random() < self.__null_probability:
            return None
        t = random.choice(schema.types)
        if self.__generate_empty_record_all_types or t == TTypeId.RECORD:
            if random.random() < self.__empty_record_probability:
                return {}

        if t == TTypeId.BIGINT:
            digits = random.random() * 10
            v = int(pow(10, digits) * random.random())
            if random.random() < .1:
                return -v
            return v
        if t == TTypeId.DOUBLE:
            digits = random.random() * 10
            v = pow(10, digits) * random.random()
            if random.random() < .1:
                return -v
            return v
        if t == TTypeId.BOOLEAN:
            if random.random() > 0.5:
                return True
            return False
        if t == TTypeId.DATE:
            return '2020-01-01'
        if t == TTypeId.TIMESTAMP_NANOS:
            return '2020-01-01 01:02:03.123'
        if t == TTypeId.STRING:
            n = random.randint(0, self.__max_string_len - 1)
            return ''.join(random.choice(string.ascii_lowercase) for i in range(n))
        if t == TTypeId.RECORD:
            return self.__generate_record(schema)
        if t == TTypeId.ARRAY:
            return self.__generate_array(schema.children['item'])
        return 'Unsupported Type %s' % t

    def __generate_array(self, schema):
        result = []
        num_children = random.randint(0, self.__max_array_len - 1)
        for _ in range(0, num_children):
            result.append(self.__generate_random_data(schema))
        return result

    def __generate_record(self, schema):
        record = {}
        for name, child in schema.children.items():
            if random.random() < self.__missing_fields_probability:
                continue
            record[name] = self.__generate_random_data(child)
        return record

    def generate(self, generate_record_idx=False):
        n = self.__min_records
        if self.__min_records != self.__max_records:
            n = random.randint(self.__min_records, self.__max_records - 1)
        data = []
        for idx in range(0, n):
            r = self.__generate_record(self.__schema)
            if generate_record_idx:
                r['idx'] = idx
            data.append(r)
        return data

class TestBase(unittest.TestCase):
    """ Base class with some common test utilities. """
    @staticmethod
    def _ddl_count(conn, sql):
        return len(conn.execute_ddl(sql))

    @staticmethod
    def _recreate_test_role(conn, role, users):
        conn.execute_ddl("DROP ROLE IF EXISTS %s" % role)
        conn.execute_ddl("CREATE ROLE %s" % role)
        for user in users:
            conn.execute_ddl("GRANT ROLE %s TO GROUP %s" % (role, user))

    @staticmethod
    def _recreate_test_db(conn, db):
        conn.execute_ddl("DROP DATABASE IF EXISTS %s CASCADE" % db)
        conn.execute_ddl("CREATE DATABASE %s" % db)

    @staticmethod
    def _create_all_types(conn, db,
                          include_date=False,
                          include_array=False,
                          include_map=False,
                          include_struct=False):
        ddl = '''
            CREATE EXTERNAL TABLE %s.alltypes(
              bool_col BOOLEAN,
              tinyint_col TINYINT,
              smallint_col SMALLINT,
              int_col INT,
              bigint_col BIGINT,
              float_col FLOAT,
              double_col DOUBLE,
              string_col STRING,
              varchar_col VARCHAR(10),
              char_col CHAR(5),
        '''
        if include_date:
            ddl += 'date_col DATE,\n'
        ddl += '''
              timestamp_col TIMESTAMP,
              decimal_col decimal(24, 10)
        '''
        if include_array:
            ddl += ',array_col ARRAY<INT>'
        if include_map:
            ddl += ',map_col MAP<STRING,INT>'
        if include_struct:
            ddl += ',struct_col STRUCT<f1:STRING,f2:INT>'

        ddl +='''
            )
            ROW FORMAT DELIMITED FIELDS TERMINATED BY '|'
            STORED AS TEXTFILE
            LOCATION 's3://cerebrodata-test/alltypes/'
        '''
        conn.execute_ddl(ddl % db)

    def _get_snowflake_data_reg_connection_obj(self, name):
        data_reg_connection = TDataRegConnection()
        data_reg_connection.name = name
        data_reg_connection.type = "JDBC"
        data_reg_connection.data_source_path = None
        data_reg_connection.jdbc_driver = "snowflake"
        data_reg_connection.host = "vq85960.snowflakecomputing.com"
        data_reg_connection.port = 0
        data_reg_connection.user_key = "awssm://snowflake-partner-username"
        data_reg_connection.password_key = "awssm://snowflake-partner-password"
        data_reg_connection.default_catalog = "SNOWFLAKE_SAMPLE_DATA"
        data_reg_connection.default_schema = "TPCH_SF1"
        data_reg_connection.is_active = True
        data_reg_connection.connection_properties = {'defaultDb':'TPCH_SF1'}
        return data_reg_connection

    @staticmethod
    def _get_t_attribute_obj(namespace, key):
        attribute = TAttribute()
        attribute.attribute_namespace = namespace
        attribute.key = key
        return attribute

    @staticmethod
    def _tag_all_columns(conn, db, dataset, ns, key, create_tag=False):
        if create_tag:
            conn.delete_attribute(ns, key)
            conn.create_attribute(ns, key)
        ds = conn.list_datasets(db, name=dataset)[0]
        for col in ds.schema.cols:
            if col.name:
                conn.assign_attribute(ns, key, db, dataset, col.name)

    @staticmethod
    def get_random_leaf_column(conn, db, tbl):
        """ Returns a random path to a leaf col in this table."""
        tbl = conn.list_datasets(db, name=tbl)[0]
        schema = TestBase.convert_schema(tbl.schema.cols)
        result = ''
        flattened = ''
        while schema:
            col = random.choice(schema)
            if col.col.name:
                if result:
                    result += '.'
                if flattened:
                    flattened += '__'
                result += col.col.name.lower()
                flattened += col.col.name.lower()
            # else:
            #     if flattened:
            #         flattened += '__'
            #     flattened += 'item'
            schema = col.children
        return result, flattened

    @staticmethod
    def convert_schema(cols):
        """ Converts a flattened inorder schema list to a tree """
        schema = []
        idx = 0
        while idx < len(cols):
            idx = SchemaNode.convert(cols, idx, schema)
        return schema

    @staticmethod
    def _visible_cols(cols):
        result = []
        for c in cols:
            if c.hidden:
                continue
            result.append(c)
        return result

    @staticmethod
    def _top_level_columns(cols):
        total_children = 0
        for c in cols:
            if c.type.num_children:
                total_children += c.type.num_children
        return len(cols) - total_children

    @staticmethod
    def collect_column_attributes(ds):
        result = {}
        for col in ds.schema.cols:
            if not col.attribute_values:
                continue
            for v in col.attribute_values:
                key = v.database
                if v.table:
                    key += '.' + v.table
                    if v.column:
                        key += '.' + v.column
                if key not in result:
                    result[key] = []
                result[key].append(
                    v.attribute.attribute_namespace + '.' + v.attribute.key)
        for _, v in result.items():
            v.sort()
        return result


    @staticmethod
    def _try_parse_datetime(v):
        if not isinstance(v, str):
            return None

        FORMATS = [
            '%Y-%m-%d %H:%M:%S',
            '%Y-%m-%d %H:%M:%S.%f',
            '%Y-%m-%dT%H:%M:%S.%fZ',
            '%Y-%m-%dT%H:%M:%S.%f',
            '%Y-%m-%dT%H:%M:%SZ',
            '%Y-%m-%dT%H:%M:%S',
            '%Y-%m-%d',
        ]

        for fmt in FORMATS:
            try:
                return datetime.datetime.strptime(v, fmt)
            except ValueError:
                pass
        return None

    @staticmethod
    def _is_float(v):
        if v is None:
            return False
        try:
            float(v)
            return True
        except:
            return False

    @staticmethod
    def _equals(v1, v2):
        if TestBase._is_float(v1) and TestBase._is_float(v2):
            v1 = float(v1)
            v2 = float(v2)

        if isinstance(v1, float) and isinstance(v2, float):
            return v1 == v2 or (math.isnan(v1) and math.isnan(v2)) \
                or abs(v1 - v2) < .0001
        if v1 != v2:
            # Try as datetime with different formats
            d1 = TestBase._try_parse_datetime(v1)
            d2 = TestBase._try_parse_datetime(v2)
            if d1 is not None and d1 == d2:
                return True
            print("Values do not match. %s != %s" % (v1, v2))
            print("Types: %s != %s" % (type(v1), type(v2)))
        return v1 == v2

    @staticmethod
    def _is_empty_dictionary(v):
        """ Checks if v is None or a dictionary (recursively) of None values """
        if v is None:
            return True
        if type(v) != dict:
            return False
        for _, val in v.items():
            if not TestBase._is_empty_dictionary(val):
                return False
        return True

    @staticmethod
    def __deep_compare(actual, expected, allow_missing, empty_struct_equals_null,
                       required_type, empty_array_equals_null, zero_equals_none):
        if empty_struct_equals_null:
            if type(actual) == dict and (expected is None or len(expected) == 0):
                # We don't support nullable structs, so instead every field in the
                # struct is None
                for k, v in actual.items():
                    if not TestBase._is_empty_dictionary(v):
                        print("actual has a non-null struct. Expecting null.")
                        return False
                return True
            if type(expected) == dict and len(expected) == 0 and actual is None:
                # Allow empty struct to be considered equal to null. A struct with no
                # fields is otherwise not possible.
                return True

        # For parquet file format, the actual is empty for an empty list.
        if empty_array_equals_null:
            if type(expected) == list and (actual is None):
                return True

        # As we are strongly typed, we will convert the types to the "bigger" type for
        # schema merge cases.
        if type(actual) == str and type(expected) in [int, float]:
            expected = str(expected)
        if type(actual) == str and type(expected) == bool:
            if expected:
                expected = "true"
            else:
                expected = "false"
        if type(actual) == float and type(expected) == int:
            expected = float(expected)

        if type(actual) != type(expected):
            if type(actual) == list and type(expected) == dict:
                # Can't tell difference between empty list and empty dict
                if not actual and not expected:
                    return True
            # Handle some schema merge cases that are ambiguous with empty records
            if type(actual) == dict and not actual and \
                    type(expected) in [bool, float, int]:
                return True
            if type(actual) in [bool, float, int] and type(expected) == dict \
                    and not expected:
                return True
            if type(actual) == float and math.isnan(actual) and expected is None:
                return True
            if actual == "" and expected is None and zero_equals_none:
                return True
            if type(actual) == type(None) and type(expected) in [float, int]:
                if zero_equals_none and expected in [0, 0.0]:
                    return True
            if required_type and type(expected) in [bool, float, int, str]:
                if type(expected) != required_type:
                    # The required type doesn't match so the expected should be None
                    if actual is not None and \
                            not TestBase._equals(actual, required_type(expected)):
                        print("Expecting actual to be None. %s, %s" % (actual, expected))
                        return False
                    return True
            print("Types don't match %s != %s" % (type(actual), type(expected)))
            print("%s != %s" % (actual, expected))
            return False

        if type(actual) == dict:
            for k, v in actual.items():
                # Handle the case where some field that exists in the Okera version
                # but is null might not exist in the raw file version, but only if
                # allow_missing is set
                if k in expected:
                    if not TestBase.__deep_compare(
                            v, expected[k], allow_missing,
                            empty_struct_equals_null, required_type,
                            empty_array_equals_null, zero_equals_none):
                        print("Key %s from expected is not in actual." % k)
                        print("%s != %s" % (v, expected[k]))
                        return False
                elif k not in expected and not \
                        (TestBase._is_empty_dictionary(v) and allow_missing):
                    print("Key %s from actual is not in expected." % k)
                    return False
            for k, v in expected.items():
                if k not in actual and not allow_missing:
                    print("Key %s from expected is not in actual." % k)
                    return False
        elif type(actual) == list:
            if len(actual) != len(expected):
                print("Length of arrays are not equal. %s != %s" % \
                      (len(actual), len(expected)))
                return False
            for idx in range(len(actual)):
                if not TestBase.__deep_compare(
                        actual[idx], expected[idx], allow_missing,
                        empty_struct_equals_null, required_type,
                        empty_array_equals_null, zero_equals_none):
                    print("List elements don't match at idx %s. %s != %s" % \
                          (idx, actual[idx], expected[idx]))
                    return False
        else:
            return TestBase._equals(actual, expected)

        return True

    @staticmethod
    def _lower_keys(x):
        if isinstance(x, list):
            return [TestBase._lower_keys(v) for v in x]
        if isinstance(x, dict):
            return dict((k.lower(), TestBase._lower_keys(v)) for k, v in x.items())
        return x

    @staticmethod
    def compare_json(asserter, actual, expected, allow_missing, empty_struct_equals_null,
                     batch_mode, required_type, empty_array_equals_null,
                     zero_equals_none=False):
        actual = TestBase._lower_keys(actual)
        expected = TestBase._lower_keys(expected)

        if batch_mode and len(actual) != len(expected):
            print("Failure: Lengths did not match %s != %s" % \
                (len(actual), len(expected)))
            return False
        asserter.assertEqual(len(actual), len(expected))

        for idx in range(len(actual)):
            obj1 = actual[idx]
            obj2 = expected[idx]

            if TestBase.__deep_compare(
                    obj1, obj2, allow_missing, empty_struct_equals_null,
                    required_type, empty_array_equals_null, zero_equals_none):
                continue
            if batch_mode:
                return False
            print("EXPECTED:\n%s" % json.dumps(expected, indent=2, sort_keys=True))
            print("\nACTUAL:\n%s" % json.dumps(actual, indent=2, sort_keys=True))
            asserter.assertEqual(json.dumps(actual, indent=2, sort_keys=True),
                                 json.dumps(expected, indent=2, sort_keys=True))
        return True

    @staticmethod
    def normalize_sql(sql):
        sql = re.sub(' +', ' ', sql)
        return " ".join(sql.split())

    def assert_sql_equals(self, expected, result, error_header=''):
        expected = self.normalize_sql(expected)
        result = self.normalize_sql(result)
        if expected != result:
            self.maxDiff = None
            msg='\n' + error_header
            msg += "--------EXPECTED---------\n"
            msg += expected + '\n'
            msg += "--------RESULT---------\n"
            msg += result + '\n'
            msg += "--------EXPECTED (formatted) ---------\n"
            msg += sqlparse.format(expected, reindent=True) + '\n'
            msg += "--------RESULT (formatted) ---------\n"
            msg += sqlparse.format(result, reindent=True) + '\n'
            print(msg)
            self.assertEqual(expected, result, msg=msg)

    def authorize_query(self, conn, query, user=None, use_tmp_tables=False,
                        client=TAuthorizeQueryClient.OKERA,
                        return_full_result=False, cte=False, token=None):
        request = TAuthorizeQueryParams()
        request.sql = query
        request.requesting_user = user
        request.use_session_local_tables = use_tmp_tables
        request.client = client
        request.cte_rewrite = cte
        if token:
            request.ctx = TRequestContext()
            request.ctx.auth_token = token

        result = conn.service.client.AuthorizeQuery(request)
        if client == TAuthorizeQueryClient.TEST_ACK:
            return ''
        if return_full_result:
            return self.normalize_sql(result.result_sql)
        self.assertTrue(result.table is None)
        if result.requires_worker:
            return None
        self.assertTrue(result.result_schema is not None or cte)
        return self.normalize_sql(result.result_sql)

    def cte_rewrite(self, conn, query, user=None,
                    generate_plan=False,
                    client=TAuthorizeQueryClient.OKERA,
                    default_db=None,
                    default_namespace=None,
                    validate_rewrite=False):
        request = TAuthorizeQueryParams()
        request.sql = query
        request.requesting_user = user
        request.cte_rewrite = True
        request.validate_rewrite = validate_rewrite
        request.plan_request = generate_plan
        request.client = client
        request.default_db = default_db
        request.default_namespace = default_namespace
        result = conn.service.client.AuthorizeQuery(request)
        if result.result_sql is None:
            return None, None
        result_sql = self.normalize_sql(result.result_sql)
        return result_sql, result.plan, result.referenced_tables, result.jdbc_referenced_tables

    def verify_tbl_access(self, conn, db, tbl, num_cols, has_db_access=False,
                            skip_metadata_check=False, timeout=0):
        """ Verifies the current connect has access to num_cols on this table

            FIXME(BUG): skip_metadata_check should be removed (and always True). It is
            skipping due to existing bugs.
        """
        if num_cols == 0:
            for ddl in ['describe %s.%s', 'describe formatted %s.%s']:
                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                    print(conn.execute_ddl(ddl % (db, tbl)))
                self.assertTrue('does not have privilege' in str(ex_ctx.exception))

            if has_db_access:
                # User has access to DB, make sure table does not show up in list
                self.assertTrue('%s.%s' %(db, tbl) not in conn.list_dataset_names(db))
                names = conn.list_dataset_names(db)
                self.assertFalse('%s.%s' %(db, tbl) in names, msg=str(names))
            else:
                # Doesn't have database access, ensure listing tables in it fails
                with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                    print("Listing datasets in: " + db)
                    print(conn.list_dataset_names(db))
                self.assertTrue('does not have privilege' in str(ex_ctx.exception))

                datasets = conn.list_datasets(db)
                self.assertEqual(len(datasets), 0, msg=str(datasets))

                for ddl in ['describe database %s', 'show tables in %s']:
                    with self.assertRaises(_thrift_api.TRecordServiceException) as ex_ctx:
                        print(conn.execute_ddl(ddl % db))
                    self.assertTrue('does not have privilege' in str(ex_ctx.exception))

                if not skip_metadata_check:
                    dbs = conn.list_databases()
                    self.assertFalse(db in dbs, msg=str(dbs))
        else:
            dbs = conn.list_databases()
            self.assertTrue(db in dbs, msg=str(dbs))
            names = conn.list_dataset_names(db)
            self.assertTrue('%s.%s' %(db, tbl) in names, msg=str(names))
            datasets = conn.list_datasets(db, name=tbl)
            self.assertEqual(len(datasets), 1)
            cols = self._visible_cols(conn.list_datasets(db, name=tbl)[0].schema.cols)
            self.assertEqual(len(cols), num_cols)

            datasets = conn.list_datasets(db)
            self.assertTrue(len(datasets) >= 1, msg=str(datasets))

            # Now do a select star and verify the right number of columns are expanded
            start = time.perf_counter()
            schema = conn.plan('SELECT * FROM %s.%s' % (db, tbl)).schema
            end = time.perf_counter()
            if schema.nested_cols:
                cols = schema.nested_cols
            else:
                cols = schema.cols
            self.assertEqual(self._top_level_columns(cols), num_cols, msg=str(cols))
            if timeout > 0:
                self.assertLess(end - start, timeout)


def random_string(length):
    return ''.join(random.choice(string.ascii_uppercase) for _ in range(length))

def create_s3_repro(conn, bucket, key):
    """ Loads a test case generated from tools/create-repro """
    import boto3
    s3 = boto3.client('s3')
    res = s3.get_object(Bucket=bucket, Key=key)
    sql = res['Body'].read().decode('UTF-8').strip(' \t\n\r')
    statements = sqlparse.split(sql)
    for stmt in statements:
        conn.execute_ddl(stmt)



from okera.tests import pycerebro_test_common as common

class MiscTest(common.TestBase):
    def test_parse_commit_message(self):
        msg = 'W3Rlc3RzXVtyZXdyaXRlXSBBZGQgc29tZSBDVEUgdGVzdHMgYW5kIHNvbWUgdGVzdCByZWZhY3RvcmluZwoKKldoYXQgdGhpcyBQUiBkb2VzIC8gd2h5IHdlIG5lZWQgaXQ6KgpGaXhlcyBvbmUgZXhwZWN0ZWQgdGVzdCBjaGFuZ2UgZnJvbToKCmh0dHBzOi8vZ2Vycml0Lm9rZXJhLmRldi9jL2NlcmVicm8vKy8xNzMwMwoKKlNwZWNpYWwgbm90ZXMgZm9yIHlvdXIgcmV2aWV3ZXI6KgoKKkFkZGl0aW9uYWwgdGVzdGluZyB0aGF0IHdhcyBkb25lOioKCkFERElUSU9OQUxfQ0lfVEVTVFM6IHNub3dmbGFrZS1jdGUKYGBgClJlbGVhc2Ugbm90ZToKYGBgCgpDaGFuZ2UtSWQ6IEk1YzBlMDRhMGM5MDUzMmNmY2RhOWMwMzQwYTQ5ZDc0ODMzOTUxMjMxCg=='
        result = common.get_ci_test_cases(msg)
        self.assertEqual(['snowflake-cte'], result)





# Copyright Okera Inc. All Rights Reserved.

import unittest

import time
from okera.tests import pycerebro_test_common as common

class JoinsTest(common.TestBase):
    @unittest.skip("This test takes a really long time to run")
    def test_join_mem(self):
        ctx = common.get_test_context()
        DB = 'joins_mem_test_db'
        with common.get_planner(ctx) as conn:
            self._recreate_test_db(conn, DB)


            conn.execute_ddl(
            """
            CREATE EXTERNAL TABLE IF NOT EXISTS %s.chase_cards_card_transaction_event_v001 (
              event_id STRING,
              event_type STRING,
              event_name STRING,
              entity_key STRING,
              partition_key STRING,
              checksum STRING,
              created_at BIGINT,
              card_transaction STRUCT<transaction_lifecycle_id:STRING,card_id:STRING,payment_device_details:STRUCT<device_id:STRING,funding_device_id:STRING,logical_card_id:STRING,device_type:STRING,token_details:STRUCT<token_number:STRING>>,party_key:STRING,state:STRING,card_messages:ARRAY<STRUCT<card_message_id:STRING,subscription_key:STRING,sequence_number:INT,matched_message_id:STRING,synthetic_trace_id:STRING,message_type:STRING,processing_category:STRING,pos_entry_mode:STRING,processing_code:STRING,acquirer_id:STRING,retrieval_reference_number:STRING,original_currency_amount:STRUCT<amount:DECIMAL(17,4),currency:STRING>,conversion_rate:DECIMAL(14,7),amount:STRUCT<amount:DECIMAL(17,4),currency:STRING>,content:STRUCT<message_type:STRING,primary_account_number:STRING,processing_code:STRING,processing_code_attributes:STRUCT<transaction_type_code:STRING,from_account_type_code:STRING,to_account_type_code:STRING>,system_trace_number:STRING,transaction:STRUCT<amounts:STRUCT<transaction:STRUCT<amount:DECIMAL(17,4),currency:STRING>,settlement:STRUCT<amount:DECIMAL(17,4),currency:STRING>,billing:STRUCT<amount:DECIMAL(17,4),currency:STRING>,fee:STRUCT<amount:DECIMAL(17,4),currency:STRING>,additional_amounts:ARRAY<STRUCT<account_type:STRING,amount_type:STRING,transaction_sign:STRING,amount:STRUCT<amount:DECIMAL(17,4),currency:STRING>>>,conversion_rate:DECIMAL(14,7)>,network_code:STRING,banknet_reference_number:STRING,dates:STRUCT<transaction_date:DATE,transaction_time:STRING,settlement_date:DATE,conversion_date:DATE,transmission_date_time:BIGINT>,retrieval_reference_number:STRING>,card:STRUCT<id:STRING,expiry_date:STRING>,acquirer:STRUCT<id:STRING,country_code:STRING>,merchant:STRUCT<name:STRING,address:STRUCT<city_name:STRING,state_or_country_code:STRING>,bankcard_phone:STRUCT<phone_number_dialed:STRING,abbreviation:STRING,call_duration:STRING,call_origin_city:STRING,call_origin_state_or_country_code:STRING>,terminal_id:STRING,category_code:STRING,acceptor_id_code:STRING>,pos:STRUCT<condition_code:STRING,additional_pos_detail:STRING,additional_pos_detail_attributes:STRUCT<pos_terminal_attendance:INT,pos_terminal_location:INT,pos_cardholder_presence:STRING,pos_card_presence:INT,pos_card_capture_capabilities:INT,pos_transaction_status:INT,pos_transaction_security:INT,cardholder_activated_terminal_level:INT,pos_card_data_terminal_input_capability_indicator:INT,pos_authorisation_life_cycle:INT,pos_country_code:STRING,pos_postal_code:STRING>,pos_entry_mode:STRING,pos_entry_mode_attributes:STRUCT<pan_entry_mode:STRING,pin_entry_mode:STRING>,extended_data_condition_codes:STRING>,additional_data:STRUCT<money_send_ref:STRING,multi_purpose_merchant_indicator:STRUCT<low_risk_merchant_indicator:STRING>,payment_initiation_channel:STRING,wallet_program_data:STRING,pan_mapping_file_information:STRING,trace_id:STRING,e_commerce_indicator:STRING,on_behalf_of_services:STRING,avs_response:STRING>,fraud_score_data:STRUCT<raw_data:STRING,way_4_risk_score:STRING,fraud_score:STRING,fraud_score_reason_code:STRING,rules_score:STRING,rules_reason_code_1:STRING,rules_reason_code_2:STRING>,wallet:STRUCT<account_number_indicator:STRING,account_number:STRING,expiry_date:STRING,token_requestor_id:STRING,wallet_program_data:STRING>,replacement_amounts:STRUCT<transaction:DECIMAL(17,4),settlement:DECIMAL(17,4),billing:DECIMAL(17,4)>,authentication:STRUCT<threeDS:STRUCT<authentication_protocol:STRING,directory_server_transaction_id:STRING>,ucaf:STRING,validationResults:STRUCT<cvc2_validation_result:STRING,ucaf_validation_result:STRING>,integratedCircuitCardData:STRING,pinValidation:STRING>,decline_reasons:ARRAY<STRUCT<reason:STRING,additional_data:MAP<STRING,STRING>>>,auth_code:STRING,auth_response_code:STRING,original_data_elements:STRUCT<message_type:STRING,system_trace_number:STRING,transmission_date_time:BIGINT,acquiring_institution_id:STRING,forwarding_institution_id:STRING>,paymentDevice:STRUCT<fundingDeviceId:STRING>,scheme_message_to_reverse_lifecycle_id:STRING,stipProcessor:STRING,stipIndicator:BOOLEAN,metadata:STRUCT<received_at:BIGINT>>,processing_result:STRUCT<response_code:STRING,auth_code:STRING,approved_amount:STRUCT<amount:DECIMAL(17,4),currency:STRING>,available_balance:STRUCT<amount:DECIMAL(17,4),currency:STRING>>,state:STRING,has_impacted_ledger:BOOLEAN,payment_device_details:STRUCT<device_id:STRING,funding_device_id:STRING,logical_card_id:STRING,device_type:STRING,token_details:STRUCT<token_number:STRING>>>>,ledger_transactions:ARRAY<STRUCT<ledger_transaction_id:STRING,subscription_key:STRING,ledger_type:STRING,original_currency_amount:STRUCT<amount:DECIMAL(17,4),currency:STRING>,average_conversion_rate:DECIMAL(16,9),amount:STRUCT<amount:DECIMAL(17,4),currency:STRING>,sign:STRING,state:STRING,transaction_code:STRING,value_date:BIGINT,booking_date:BIGINT,correlation_id:STRING>>,clearing_message:STRUCT<clearing_message_id:STRING,cleared_original_currency_amount:STRUCT<amount:DECIMAL(17,4),currency:STRING>,conversion_rate:DECIMAL(16,9),cleared_amount:STRUCT<amount:DECIMAL(17,4),currency:STRING>,acquirer_reference_data:STRING,content:STRUCT<id:STRING,transaction_date:STRING,retrieval_reference_number:STRING,merchant:STRUCT<name:STRING,address:STRUCT<city_name:STRING,state_code:STRING,country_code:STRING,post_code:STRING>,terminal_id:STRING,category_code:STRING,acceptor_id_code:STRING>,card:STRUCT<id:STRING,expiry_date:STRING,card_sequence_number:INT,primary_account_number:STRING>,amounts:STRUCT<transaction:STRUCT<amount:DECIMAL(17,4),currency:STRING>,settlement:STRUCT<amount:DECIMAL(17,4),currency:STRING>,billing:STRUCT<amount:DECIMAL(17,4),currency:STRING>,transaction_sign:STRING,interchange_fee:STRUCT<amount:DECIMAL(17,4),currency:STRING>,interchange_fee_sign:STRING,additional_amounts:ARRAY<STRUCT<amount_type:STRING,amount:STRUCT<amount:DECIMAL(17,4),currency:STRING>>>,settlement_conversion_rate:DECIMAL(16,9),billing_conversion_rate:DECIMAL(16,9)>,authentication_code:STRING,mapped_authorisation_processing_code:STRING,processing_code:STRING,pos:STRUCT<message_reason_code:STRING,pos_entry_mode:STRING>,function_code:INT,transaction_lifecycle_id:STRING,icc_data:STRING,ecommerce_security_level_indicator:STRING,mapping_service_account_number:STRING,wallet_identifier:STRING,data_record:STRING,message_number:INT,matching_indicator:BOOLEAN,file_id:STRING,acquirer_reference_data:STRING,created_at:BIGINT,file_key:STRING,row_number:INT,paymentDevice:STRUCT<fundingDeviceId:STRING>,wallet:STRUCT<mapping_service_account_number:STRING,wallet_identifier:STRING,token_requestor_id:STRING>>>,version:INT>
                          )
            PARTITIONED BY (
              dt DATE
              )
            COMMENT 'Discovered by Okera crawler'
            WITH SERDEPROPERTIES ('avro.schema.url'='s3a://cerebrodata-test/chase/chase.cards.card-transaction-event-v001/dt=2022-01-13/chase.cards.card-transaction-event-v001 34 0000008419.avro')
            STORED AS AVRO
            LOCATION 's3a://cerebrodata-test/chase/chase.cards.card-transaction-event-v001'
            TBLPROPERTIES ('okera.view.children'='chase_payments_view.chase_cards_card_transaction_event_view')
            """ % DB)

            conn.execute_ddl("""
                ALTER TABLE %s.chase_cards_card_transaction_event_v001 RECOVER PARTITIONS
            """ % DB)

            result = conn.scan_as_json("""
              select * from %s.chase_cards_card_transaction_event_v001
                  t, t.card_transaction.card_messages m
              WHERE m.item.content.wallet.wallet_program_data = '103'""" % DB,
              max_task_count=1)
            print(result)

            result = conn.scan_as_json("""
              select * from %s.chase_cards_card_transaction_event_v001
                  t, t.card_transaction.card_messages m""" % DB, max_task_count=1)
            print(result)

    @unittest.skip("This test requires --fast-sasl-timeout to start the cluster")
    def test_client_sleep(self):
        ctx = common.get_test_context(auth_mech='TOKEN')
        DB = 'joins_mem_test_db'
        with common.get_planner(ctx) as conn:
            print('Creating db...')
            self._recreate_test_db(conn, DB)
            print('Sleeping...')
            time.sleep(30)
            print('Creating db...')
            self._recreate_test_db(conn, DB)

if __name__ == "__main__":
    unittest.main()
